# -*- coding: utf-8 -*-
# (c) 2019 Richard Pobering <richard@hiveeyes.org>
# (c) 2019 Andreas Motl <andreas@hiveeyes.org>
# License: GNU General Public License, Version 3

from .core import NetworkManager
from .wifi import SystemWiFiMetrics, WiFiException



# -*- coding: utf-8 -*-
# (c) 2020 Richard Pobering <richard@hiveeyes.org>
# (c) 2020 Andreas Motl <andreas@hiveeyes.org>
# License: GNU General Public License, Version 3
import sys


def monkeypatch_cpython():

    # Setup logging.
    monkeypatch_logging()

    # Reconfigure other libraries.
    monkeypatch_stdlib()
    monkeypatch_exceptions()
    monkeypatch_machine()
    monkeypatch_pycom()


def monkeypatch_stdlib():

    from mock import Mock

    import time
    sys.modules['utime'] = time

    import builtins
    builtins.const = int

    import struct
    sys.modules['ustruct'] = struct

    import binascii
    sys.modules['ubinascii'] = binascii

    import time
    def ticks_ms():
        import time
        return time.time() * 1000
    def ticks_diff(ticks1, ticks2):
        return abs(ticks1 - ticks2)
    time.ticks_ms = ticks_ms
    time.ticks_diff = ticks_diff
    sys.modules['utime'] = time

    sys.modules['micropython'] = Mock()
    sys.modules['micropython'].const = int

    import io
    sys.modules['uio'] = io

    import os
    sys.modules['uos'] = os

    import ssl
    sys.modules['ussl'] = ssl

    import socket
    sys.modules['usocket'] = socket

    import requests
    sys.modules['urequests'] = requests

    import gc
    import psutil
    gc.threshold = Mock()
    gc.mem_free = lambda: psutil.virtual_memory().available * 100 / psutil.virtual_memory().total
    gc.mem_alloc = lambda: psutil.virtual_memory().percent

    # Optional convenience to improve speed.
    gc.collect = Mock()


def monkeypatch_exceptions():

    import traceback

    def print_exception(exc, file=sys.stdout):
        traceback.print_exception(exc, exc, exc.__traceback__)

    sys.print_exception = print_exception


def monkeypatch_machine():

    from mock import Mock

    import uuid
    import machine
    import psutil

    # Some primitives.
    machine.enable_irq = Mock()
    machine.disable_irq = Mock()
    machine.unique_id = lambda: str(uuid.uuid4().fields[-1])[:5].encode()
    machine.freq = lambda: psutil.cpu_freq().current * 1000000
    machine.idle = Mock()

    # Reset cause and wake reason.
    machine.PWRON_RESET = 0
    machine.HARD_RESET = 1
    machine.WDT_RESET = 2
    machine.DEEPSLEEP_RESET = 3
    machine.SOFT_RESET = 4
    machine.BROWN_OUT_RESET = 5

    machine.PWRON_WAKE = 0
    machine.GPIO_WAKE = 1
    machine.RTC_WAKE = 2
    machine.ULP_WAKE = 3

    machine.reset_cause = Mock(return_value=0)
    machine.wake_reason = wake_reason


def wake_reason():
    from terkin.util import get_platform_info
    platform_info = get_platform_info()
    if platform_info.vendor == platform_info.MICROPYTHON.Pycom:
        return 0, None
    else:
        return 0


def monkeypatch_pycom():

    from mock import MagicMock

    import network
    network.Bluetooth = MagicMock()
    network.LTE = MagicMock()

    sys.modules['pycom'] = MagicMock()


def monkeypatch_logging():

    import io
    import logging

    def exc(self, e, msg, *args):
        buf = io.StringIO()
        sys.print_exception(e, buf)
        self.log(logging.ERROR, msg + "\n" + buf.getvalue(), *args)

    logging.Logger.exc = exc



# -*- coding: utf-8 -*-
# (c) 2005,2020 Andreas Motl <andreas@hiveeyes.org>
# (c) 2005,2020 Jan Hoffmann <jan.hoffmann@bergamsee.de>
# License: GNU General Public License, Version 3
import re
import json
import serial
import struct


class EPSolar:

    def __init__(self, device):
        self.config = device
        self.serial_data = []
        self.serial_params = '''
            {
                "serial-port": "''' + str(device) + '''",
                "serial-params": {
                    "bytesize": 8,
                    "baudrate": 9600,
                    "timeout": 2.0,
                    "parity": "N"
                    }
            }
            '''

        self.config = json.loads(self.serial_params)
        self.config['ser'] = serial.Serial(str(self.config.get('serial-port')), **self.config.get('serial-params'))

    def read_serial(self):
        string = b'\xAA\x55\xEB\x90\xEB\x90\xEB\x90\x16\xA0\x00\xB1\xA7\x7F'
        self.config.get('ser').write(string)
        payload = self.config.get('ser').readline()
        # 1. decode payload
        try:
            data_length = payload[8]
            pos_data_start = 9
            pos_data_stop = pos_data_start + data_length

            data = payload[pos_data_start:pos_data_stop]

        except Exception as exc:
            # print traceback.format_exc()
            print(exc)
            # self.output(exc)
            serial_read = False
            return False, serial_read

        # 2. decode data
        # http://docs.python.org/2/library/struct.html#format-characters
        try:
            serial_data = list(struct.unpack('hhhhhh???c????bh', data))
            serial_read = True
            # 15 Battery temp., there is 30 difference in value with the real value (spec)
            serial_data[14] -= 35
            # 35 seems to be the right value
            self.serial_data = serial_data
            return serial_data, serial_read

        except Exception as exc:
            serial_read = False
            return False, serial_read

    def decode_serial(self, payload):
        if payload[1] < 100:
            payload[1] = 0
        if isinstance(payload[0], str):
            payload[0] = self.compile_value(payload[0], 'volt')
        if isinstance(payload[1], str):
            payload[1] = self.compile_value(payload[1], 'volt')
        if isinstance(payload[3], str):
            payload[3] = self.compile_value(payload[3], 'ampere')
        if isinstance(payload[6], str):
            payload[6] = self.compile_value(payload[6], 'ampere')
        if isinstance(payload[9], str):
            try:
                payload[9] = float(payload[9])
            except Exception as exc:
                payload[9] = float(0)

        return payload

    def compile_value(self, value, unit):
        if 2 > 1:
            # try:
            if unit == "ampere":
                p = re.compile(r'^(.)(..)*')
            if unit == "volt":
                p = re.compile(r'^(..)(..)*')
            m = re.match(p, value)
            if m.group(1):
                int1 = m.group(1)
            else:
                int1 = 0
            if m.group(2):
                int2 = m.group(2)
            else:
                int2 = 0
            result = str(int1) + "." + str(int2)
            # except:
            # result = "False"
        return result

    def serial_data_prepare(self, payload):
        data = dict(
            battery_voltage=float(payload[0]/100),
            pv_voltage=float(payload[1]/100),
            reserved1=str(payload[2]),
            load_current=float(payload[3]),
            over_discharge_voltage=float(payload[4]/100),
            battery_full_voltage=float(payload[5]/100),
            load=str(payload[6]),
            overload=str(payload[7]),
            load_short_circuit=str(payload[8]),
            # reserved2=str(payload[9]),
            battery_overload=str(payload[10]),
            over_discharge=str(payload[11]),
            full_indicator=str(payload[12]),
            charging_indicator=str(payload[13]),
            battery_temperature=float(payload[14]),
            charging_current=str(payload[15]))

        for key, value in data.items():
            
            try:
                if self.isnumeric(value):
                    data[key] = float(value)
            except Exception as exc:
                pass

        return data


if __name__ == "__main__":
    config = ()
    epsolar = Epsolar(config)

    # Connect serial port
    serial_data, serial_read = epsolar.read_serial()
    print('Serial Data: ', format(serial_data))

    serial_data_converted = epsolar.decode_serial(serial_data)
    serial_data_decoded = epsolar.serial_data_prepare(serial_data_converted)
    serial_data_prepared = epsolar.serial_data_prepare(serial_data)

    print(serial_data_prepared)
    print(serial_data_decoded)



# -*- coding: utf-8 -*-
# (c) 2017-2019 Andreas Motl <andreas@terkin.org>
# (c) 2019 Richard Pobering <richard@hiveeyes.org>
# License: GNU General Public License, Version 3
import os
import sys
import time

import machine

from umal import ApplicationInfo
from terkin import logging
from terkin.telemetry.core import TelemetryManager, TelemetryAdapter
from terkin.util import get_device_id
from terkin.watchdog import Watchdog

log = logging.getLogger(__name__)


class DeviceStatus:
    """Object holding device status information."""
    def __init__(self):
        self.maintenance = False
        self.networking = False


class TerkinDevice:
    """
    Singleton object for enabling different device-related subsystems
    and providing lowlevel routines for sleep/resume functionality.
    """

    def __init__(self, application_info: ApplicationInfo):

        self.application_info = application_info
        self.platform_info = application_info.platform_info

        self.name = application_info.name
        self.version = application_info.version
        self.settings = application_info.settings

        self.status = DeviceStatus()
        self.watchdog = Watchdog(device=self, settings=self.settings)

        # Conditionally enable terminal on UART0. Default: False.
        #try:
        #    self.terminal = Terminal(self.settings)
        #    self.terminal.start()
        #except Exception as ex:
        #    log.exc(ex, 'Enabling Terminal failed')

        self.device_id = get_device_id()

        self.networking = None
        self.telemetry = None

        self.rtc = None

    def start_networking(self):
        """ 
        Start all configured networking devices.
        """
        log.info('Starting networking')

        from terkin.network import NetworkManager, WiFiException

        self.networking = NetworkManager(device=self, settings=self.settings)

        if self.settings.get('networking.wifi.enabled'):
            # Start WiFi.
            try:
                self.networking.start_wifi()

            except Exception as ex:
                log.exc(ex, 'Starting WiFi networking failed')
                self.status.networking = False
                return

            # Wait for network stack to come up.
            try:
                self.networking.wait_for_ip_stack(timeout=5)
                self.status.networking = True
            except Exception as ex:
                log.exc(ex, 'IP stack not available')
                self.status.networking = False

        else:
            log.info("[WiFi] Interface not enabled in settings.")

        try:
            self.networking.start_services()
        except Exception as ex:
            log.exc(ex, 'Starting network services failed')

        # Initialize LoRa device.
        platform_info = self.application_info.platform_info
        is_pycom_lora = platform_info.device_name in ['LoPy', 'LoPy4', 'FiPy']
        is_dragino = platform_info.vendor == platform_info.MICROPYTHON.RaspberryPi
        if self.settings.get('networking.lora.enabled'):
            if is_pycom_lora or is_dragino:
                if self.settings.get('networking.lora.antenna_attached'):
                    try:
                        self.networking.start_lora()
                        self.status.networking = True
                    except Exception as ex:
                        log.exc(ex, 'Unable to start LoRa subsystem')
                        self.status.networking = False
                else:
                    log.info("[LoRa] Disabling LoRa interface as no antenna has been attached. "
                                 "ATTENTION: Running LoRa without antenna will wreck your device.")
            else:
                log.error("[LoRa] This is not a LoRa capable device.")
        else:
            log.info("[LoRa] Interface not enabled in settings.")

        # Initialize LTE modem.
        if self.settings.get('networking.lte.enabled'):
            try:
                self.networking.start_lte()
                self.status.networking = True
            except Exception as ex:
                log.exc(ex, 'Unable to start LTE modem')
                self.status.networking = False
        else:
            log.info("[LTE]  Interface not enabled in settings.")

        # Initialize GPRS modem.
        if self.settings.get('networking.gprs.enabled'):
            try:
                self.networking.start_gprs()
                self.status.networking = True
            except Exception as ex:
                log.exc(ex, 'Unable to start GPRS modem')
                self.status.networking = False
        else:
            log.info("[GPRS] Interface not enabled in settings.")

        # Inform about networking status.
        #self.networking.print_status()

    def start_rtc(self):
        """
        The RTC is used to keep track of the date and time.
        Syncs RTC with a NTP server.
        """
        # https://docs.pycom.io/firmwareapi/pycom/machine/rtc.html
        # https://medium.com/@chrismisztur/pycom-uasyncio-installation-94931fc71283
        import time
        from machine import RTC
        self.rtc = RTC()
        # TODO: Use values from configuration settings here.
        self.rtc.ntp_sync("pool.ntp.org", 360)
        while not self.rtc.synced():
            time.sleep_ms(50)
        log.info('RTC: %s', self.rtc.now())

    def run_gc(self):
        """
        Curate the garbage collector.
        https://docs.pycom.io/firmwareapi/micropython/gc.html

        For a "quick fix", issue the following periodically.
        https://community.hiveeyes.org/t/timing-things-on-micropython-for-esp32/2329/9

        """
        import gc
        log.info('Start curating the garbage collector')
        gc.threshold(gc.mem_free() // 4 + gc.mem_alloc())
        log.info('Collecting garbage')
        gc.collect()
        #log.info('Curating the garbage collector finished')
        log.info('Curating the garbage collector finished. Free memory: %s', gc.mem_free())

    def configure_rgb_led(self):
        """https://docs.pycom.io/tutorials/all/rgbled.html"""
        if self.platform_info.vendor == self.platform_info.MICROPYTHON.Pycom:
            import pycom
            # Enable or disable heartbeat.
            rgb_led_heartbeat = self.settings.get('main.rgb_led.heartbeat', True)
            terkin_blink_pattern = self.settings.get('main.rgb_led.terkin', False)
            if terkin_blink_pattern:
                rgb_led_heartbeat = False
            pycom.heartbeat(rgb_led_heartbeat)
            pycom.heartbeat_on_boot(rgb_led_heartbeat)

    def blink_led(self, color, count=1):
        """

        :param color: rgb value as three hex values 0-255, e.g. 0x00FF00 for green
        :param count:  (Default value = 1)

        """
        if self.platform_info.vendor == self.platform_info.MICROPYTHON.Pycom:
            import pycom
            terkin_blink_pattern = self.settings.get('main.rgb_led.terkin', False)
            if terkin_blink_pattern:
                for _ in range(count):
                    pycom.rgbled(color)
                    time.sleep(0.15)
                    pycom.rgbled(0x000000)
                    time.sleep(0.10)
        elif self.settings.get('main.rgb_led.simple', False):
            led = machine.Pin(int(self.settings.get('main.rgb_led.pin')[1:]),machine.Pin.OUT)
            for _ in range(count):
                led.value(1)
                time.sleep(0.15)
                led.value(0)
                time.sleep(0.10)
            


    def start_telemetry(self):
        """ """
        log.info('Starting telemetry')

        self.telemetry = TelemetryManager()

        # Read all designated telemetry targets from configuration settings.
        telemetry_targets = self.settings.get('telemetry.targets', [])

        # Compute list of all _enabled_ telemetry targets.
        telemetry_candidates = []
        for telemetry_target in telemetry_targets:
            if telemetry_target.get('enabled', False):
                telemetry_candidates.append(telemetry_target)

        # Create adapter objects for each enabled telemetry target.
        for telemetry_target in telemetry_candidates:
            try:
                self.create_telemetry_adapter(telemetry_target)
                self.watchdog.feed()

            except Exception as ex:
                log.exc(ex, 'Creating telemetry adapter failed for target: %s', telemetry_target)

    def create_telemetry_adapter(self, telemetry_target):
        """

        :param telemetry_target: 

        """
        # Create adapter object.
        telemetry_adapter = TelemetryAdapter(device=self, target=telemetry_target)

        # Setup telemetry adapter.
        telemetry_adapter.setup()

        self.telemetry.add_adapter(telemetry_adapter)

    def enable_serial(self):
        """ """
        # Disable these two lines if you don't want serial access.
        # The Pycom forum tells us that this is already incorporated into
        # more recent firmwares, so this is probably a thing of the past.
        #uart = machine.UART(0, 115200)
        #os.dupterm(uart)
        pass

    def print_bootscreen(self):
        """Print bootscreen.
        
        This contains important details about your device
        and the operating system running on it.


        """

        if not self.settings.get('main.logging.enabled', False):
            return

        # Todo: Maybe refactor to TerkinDatalogger.
        from uio import StringIO
        buffer = StringIO()

        def add(item=''):
            """

            :param item:  (Default value = '')

            """
            buffer.write(item)
            buffer.write('\n')

        # Program name and version.
        title = '{} {}'.format(self.name, self.version)

        add()
        add('=' * len(title))
        add(title)
        add('=' * len(title))

        # Machine runtime information.
        frequency = machine.freq() / 1000000

        add('Device id    {}'.format(self.device_id))
        add()
        add('CPU freq     {}   MHz'.format(frequency))
        try:
            import pycom
            free_heap = pycom.get_free_heap()
            add('{:13}{:>7} {}'.format('Free heap', free_heap[0] / 1000.0, 'kB'))
            add('{:13}{:>7} {}'.format('Free PSRAM', free_heap[1] / 1000.0, 'kB'))
        except:
            pass
        add()

        # System memory info (in bytes).
        """
        if hasattr(machine, 'info'):
            machine.info()
            add()
        """

        # TODO: Python runtime information.
        add('{:8}: {}'.format('Python', sys.version.replace('\n', '')))
        add('{:8}: {}'.format('platform', sys.platform))

        """
        >>> import os; os.uname()
        (sysname='FiPy', nodename='FiPy', release='1.20.0.rc7', version='v1.9.4-2833cf5 on 2019-02-08', machine='FiPy with ESP32', lorawan='1.0.2', sigfox='1.0.1')
        """
        runtime_info = os.uname()
        #print(dir(runtime_info))
        for key in dir(runtime_info):
            if key.startswith('__') or key.startswith('n_'):
                continue
            value = getattr(runtime_info, key)
            if callable(value):
                continue
            #print('value:', value)
            add('{:8}: {}'.format(key, value))
        add()

        # Todo: Add program authors, contributors and credits.

        log.info('\n' + buffer.getvalue())

    def power_off_lte_modem(self):
        """
        We don't use LTE yet.

        Important
        =========
        Once the LTE radio is initialized, it must be de-initialized
        before going to deepsleep in order to ensure minimum power consumption.
        This is required due to the LTE radio being powered independently and
        allowing use cases which require the system to be taken out from
        deepsleep by an event from the LTE network (data or SMS received for
        instance).

        Note
        ====
        When using the expansion board and the FiPy together, the RTS/CTS
        jumpers MUST be removed as those pins are being used by the LTE radio.
        Keeping those jumpers in place will lead to erratic operation and
        higher current consumption specially while in deepsleep.

        -- https://forum.pycom.io/topic/3090/fipy-current-consumption-analysis/17

        See also
        ========
        - https://community.hiveeyes.org/t/lte-modem-des-pycom-fipy-komplett-stilllegen/2161
        - https://forum.pycom.io/topic/4877/deepsleep-on-batteries/10
        """

        log.info('Turning off LTE modem')
        try:
            import pycom
            from network import LTE

            log.info('Turning off LTE modem on boot')
            pycom.lte_modem_en_on_boot(False)

            # Disables LTE modem completely. This presumably reduces the power
            # consumption to the minimum. Call this before entering deepsleep.
            log.info('Invoking LTE.deinit()')
            lte = LTE()
            lte.deinit(detach=False, reset=True)

        except Exception as ex:
            log.exc(ex, 'Shutting down LTE modem failed')

    def power_off_bluetooth(self):
        """We don't use Bluetooth yet."""

        if self.platform_info.vendor == self.platform_info.MICROPYTHON.Vanilla:
            log.warning("FIXME: Skip touching Bluetooth on vanilla MicroPython "
                        "platforms as we don't use Bluetooth yet")
            return

        log.info('Turning off Bluetooth')
        try:
            from network import Bluetooth
            bluetooth = Bluetooth()
            bluetooth.deinit()
        except Exception as ex:
            log.exc(ex, 'Shutting down Bluetooth failed')

    def hibernate(self, interval, lightsleep=False, deepsleep=False):
        """

        :param interval:
        :param lightsleep:  (Default value = False)
        :param deepsleep:  (Default value = False)

        """

        #logging.enable_logging()

        if deepsleep:

            # Prepare and invoke deep sleep.
            # https://docs.micropython.org/en/latest/library/machine.html#machine.deepsleep

            log.info('Preparing deep sleep')

            # Set wake up mode.
            self.set_wakeup_mode()

            # Invoke deep sleep.
            log.info('Entering deep sleep for {} seconds'.format(interval))
            #self.terminal.stop()
            machine.deepsleep(int(interval * 1000))

        else:

            # Adjust watchdog for interval.
            self.watchdog.adjust_for_interval(interval)

            # Invoke light sleep.
            # https://docs.micropython.org/en/latest/library/machine.html#machine.sleep
            # https://docs.micropython.org/en/latest/library/machine.html#machine.lightsleep
            #
            # As "machine.sleep" seems to be a noop on Pycom MicroPython,
            # we will just use the regular "time.sleep" here.
            # machine.sleep(int(interval * 1000))
            machine.idle()

            if lightsleep:
                log.info('Entering light sleep for {} seconds'.format(interval))
                machine.sleep(int(interval * 1000))

            else:
                # Normal wait.
                log.info('Waiting for {} seconds'.format(interval))
                time.sleep(interval)

    def resume(self):
        """ """
        try:
            from terkin.pycom import MachineResetCause
            log.info('Reset cause and wakeup reason: %s', MachineResetCause().humanize())
        except Exception as ex:
            log.exc(ex, 'Could not determine reset cause and wakeup reason')

    def set_wakeup_mode(self):
        """ """

        # Set wake up parameters.
        """
        The arguments are:

        - pins: a list or tuple containing the GPIO to setup for deepsleep wakeup.

        - mode: selects the way the configured GPIOs can wake up the module.
          The possible values are: machine.WAKEUP_ALL_LOW and machine.WAKEUP_ANY_HIGH.

        - enable_pull: if set to True keeps the pull up or pull down resistors enabled
          during deep sleep. If this variable is set to True, then ULP or capacitive touch
          wakeup cannot be used in combination with GPIO wakeup.

        -- https://community.hiveeyes.org/t/deep-sleep-with-fipy-esp32-on-micropython/1792/12

        This will yield a wake up reason like::

            'wakeup_reason': {'code': 1, 'message': 'PIN'}

        """

        # Todo: ``enable_pull`` or not?

        # From documentation.
        # machine.pin_sleep_wakeup(pins=['P8'], mode=machine.WAKEUP_ALL_LOW, enable_pull=True)

        # Let's try.
        #log.info('Configuring Pin 4 for wakeup from deep sleep')
        #machine.pin_sleep_wakeup(pins=['P4'], mode=machine.WAKEUP_ALL_LOW, enable_pull=True)
        #machine.pin_sleep_wakeup(pins=['P4'], mode=machine.WAKEUP_ANY_HIGH, enable_pull=True)
        pass


class Terminal:
    """ """

    def __init__(self, settings):
        self.settings = settings
        self.uart = None

    def start(self):
        """Start Terminal on UART0 interface."""
        # Conditionally enable terminal on UART0. Default: False.
        # https://forum.pycom.io/topic/1224/disable-console-to-uart0-to-use-uart0-for-other-purposes
        uart0_enabled = self.settings.get('interfaces.uart0.terminal', False)
        if uart0_enabled:
            from machine import UART
            self.uart = UART(0, 115200)
            #self.uart = UART(0)
            os.dupterm(self.uart)
        else:
            self.shutdown()

    def stop(self):
        """Shut down."""
        log.info('Shutting down Terminal')
        self.shutdown()

    def shutdown(self):
        """Shut down Terminal and UART0 interface."""
        os.dupterm(None)
        self.deinit()

    def deinit(self):
        """Shut down UART0 interface."""
        if self.uart:
            log.info('Shutting down UART0')
            self.uart.deinit()



# -*- coding: utf-8 -*-
# (c) 2019 Richard Pobering <richard@hiveeyes.org>
# (c) 2020 Jan Hoffmann <jan.hoffmann@bergamsee.de>
# (c) 2019 Andreas Motl <andreas@hiveeyes.org>
# License: GNU General Public License, Version 3
from terkin import logging
from terkin.sensor import SensorManager, AbstractSensor
from terkin.util import get_platform_info

log = logging.getLogger(__name__)
platform_info = get_platform_info()

def includeme(sensor_manager: SensorManager, sensor_info):
    """
    Create Gpsd sensor object.

    :param sensor_manager:
    :param sensor_info:

    :return: sensor_object
    """
    sensor_object = GpsdSensor(settings=sensor_info)
    return sensor_object

class GpsdSensor(AbstractSensor):
    """
    A generic GPSD sensor component.
    Used by terkin/datalogger to register and read() from this sensor.
    start() & read() are mandatory.
    """

    def __init__(self, settings=None):

        super().__init__(settings=settings)

        # Can be overwritten by ``.set_address()``.
        self.address = None
        self.driver = None

    def start(self):
        """Getting the bus"""

        log.info('Initializing sensor GPSD')

        # Initialize the hardware driver.
        try:

            # Vanilla MicroPython 1.11 and Pycom MicroPython 1.9.4
            if platform_info.vendor in [platform_info.MICROPYTHON.Vanilla, platform_info.MICROPYTHON.Pycom]:
                raise NotImplementedError('No GPSD on MicroPython')

            # CPython SerialBus EPSolar
            elif platform_info.vendor == platform_info.MICROPYTHON.RaspberryPi:
                from terkin.lib.gpsd import Gpsd
                self.driver = Gpsd()

            else:
                raise NotImplementedError('GPSD driver not implemented on this platform')

            return True

        except Exception as ex:
            log.exc(ex, 'GPSD hardware driver failed')

    def read(self):
        """ """
        log.info('Reading sensor GPSD')

        if self.driver is None:
            return self.SENSOR_NOT_INITIALIZED

        # Vanilla MicroPython 1.11 and Pycom MicroPython 1.9.4
        if platform_info.vendor in [platform_info.MICROPYTHON.Vanilla, platform_info.MICROPYTHON.Pycom]:
            raise NotImplementedError('No GPSD on MicroPython')

        elif platform_info.vendor == platform_info.MICROPYTHON.RaspberryPi:
            data = self.driver.read()
            d = {}
            for key, value in data.items():
                key = '{}.gpsd'.format(key)
                d[key] = value
            return d



# -*- coding: utf-8 -*-
# (c) 2019-2020 Jan Hoffmann <jan.hoffmann@bergamsee.de>
# (c) 2019-2020 Andreas Motl <andreas@hiveeyes.org>
# License: GNU General Public License, Version 3
from terkin import logging
from terkin.model import SensorReading
from terkin.sensor import SensorManager, AbstractSensor
from terkin.util import get_platform_info

log = logging.getLogger(__name__)
platform_info = get_platform_info()


def includeme(sensor_manager: SensorManager, sensor_info):
    """
    Create BMP280 sensor object.

    :param sensor_manager:
    :param sensor_info:

    :return: sensor_object
    """
    sensor_object = BMP280Sensor(settings=sensor_info)
    sensor_bus = sensor_manager.get_bus_by_name(sensor_info.get('bus'))
    sensor_object.acquire_bus(sensor_bus)

    return sensor_object


class BMP280Sensor(AbstractSensor):
    """
    A generic BMP280 sensor component.
    Used by terkin/datalogger to register and read() from this sensor.
    start() & read() are mandatory.
    """

    def __init__(self, settings=None):

        super().__init__(settings=settings)

        # Can be overwritten by ``.set_address()``.
        self.address = self.settings.get('address', 0x76)

    def start(self):
        """
        Setup the BMP280 sensor driver.

        :return:
        """

        # Ensure a bus object exists and is ready.
        self.ensure_bus()

        # Initialize the hardware driver.
        try:

            # MicroPython
            # https://github.com/dafvid/micropython-bmp280

            # Adafruit CircuitPython
            if platform_info.vendor == platform_info.MICROPYTHON.RaspberryPi:
                import adafruit_bmp280
                self.driver = adafruit_bmp280.Adafruit_BMP280_I2C(i2c=self.bus.adapter, address=self.address)

            else:
                raise NotImplementedError('BMP280 driver not implemented on this platform')

            return True

        except Exception as ex:
            log.exc(ex, 'BMP280 hardware driver failed')
            return False

    def read(self):
        """
        Read the BMP280 sensor.

        :return: SensorReading
        """

        if self.bus is None or self.driver is None:
            return self.SENSOR_NOT_INITIALIZED

        #log.info('Acquire reading from BMP280')

        data = {}

        # Adafruit CircuitPython
        if platform_info.vendor == platform_info.MICROPYTHON.RaspberryPi:

            values = {
                "temperature": self.driver.temperature,
                "pressure": self.driver.pressure,
            }

        # Build telemetry payload.
        # TODO: Push this further into the telemetry domain.
        fieldnames = values.keys()
        for name in fieldnames:
            fieldname = self.format_fieldname(name, hex(self.address))
            value = values[name]
            data[fieldname] = value

        if not data:
            log.warning("I2C device {} has no value: {}".format(self.address, data))

        log.debug("I2C data:     {}".format(data))

        reading = SensorReading()
        reading.sensor = self
        reading.data = data

        return reading



# -*- coding: utf-8 -*-
# (c) 2019 Richard Pobering <richard@hiveeyes.org>
# (c) 2019 Andreas Motl <andreas@hiveeyes.org>
# License: GNU General Public License, Version 3
from terkin import logging
from terkin.sensor import AbstractSensor

log = logging.getLogger(__name__)


class ADCSensor(AbstractSensor):

    def __init__(self):

        # Hardware parameters and configuration settings.

        super().__init__()

        # The driver instance.
        self.adc = None
        self.channel = None


    def start(self):
        try:
            from machine import ADC
            adc = ADC()
            self.channel = adc.channel(pin=self.pins['adc_in'], attn=ADC.ATTN_11DB)
            #p_out = Pin('P9', mode=Pin.OUT, pull=Pin.PULL_DOWN)

        except Exception as ex:
            log.exception('ADC hardware driver failed')
            raise

    def read_value(self):
        if self.channel is None:
            # TODO: Return Sensor.DISABLED
            return

        #log.info('Acquire reading from ADC')
        value = self.channel.value()

        return value


class MoistureSensor(ADCSensor):

    def read(self):
        value = self.read_value()
        data = {
            "moisture_raw": value,
            "moisture_volt": value / self.parameter.get('scaling', 4.096),
        }
        return data


class WaterlevelSensor(ADCSensor):

    def read(self):
        value = self.read_value()
        data = {
            "waterlevel_raw": value,
            "waterlevel_volt": value / self.parameter.get('scaling', 4.096),
        }
        return data



# -*- coding: utf-8 -*-
# (c) 2019-2020 Andreas Motl <andreas@hiveeyes.org>
# (c) 2019-2020 Richard Pobering <richard@hiveeyes.org>
# (c) 2019-2020 Jan Hoffmann <jan.hoffmann@bergamsee.de>
# License: GNU General Public License, Version 3
import time


class SensorReading:
    """
    Object for holding a sensor reading.
    """
    def __init__(self):

        # Reference to the sensor object.
        self.sensor = None

        # Data dictionary of sensor values.
        self.data = None

        # Time when the reading has been taken.
        self.timestamp = time.time()


class DataFrame:
    """
    A collection of sensor readings to be send from device.
    """
    def __init__(self):

        # List of SensorReading objects.
        self.readings = []

        # Data dictionary of all sensor values.
        self.data_in = {}

        # Transformed data dictionary of all sensor values.
        self.data_out = {}

        # Serialized telemetry payload for all sensor values.
        self.payload_out = None



# -*- coding: utf-8 -*-
# (c) 2019 Richard Pobering <richard@hiveeyes.org>
# (c) 2019 Andreas Motl <andreas@hiveeyes.org>
# License: GNU General Public License, Version 3
import time
import socket
from terkin import logging
from terkin.util import get_platform_info

log = logging.getLogger(__name__)
platform_info = get_platform_info()


class LoRaAdapter:

    def __init__(self, network_manager, settings):
        self.network_manager = network_manager
        self.settings = settings

        # Pycom MicroPython
        if platform_info.vendor in [platform_info.MICROPYTHON.Vanilla, platform_info.MICROPYTHON.Pycom]:
            self.driver = LoRaDriverPycom(self.network_manager, self.settings)

        # RaspberryPi (Dragino)
        elif platform_info.vendor == platform_info.MICROPYTHON.RaspberryPi:
            self.driver = LoRaDriverDragino(self.network_manager, self.settings)

    def start(self):
        log.info('[LoRa] Starting LoRa Adapter')
        return self.driver.start()

    def ensure_connectivity(self):
        return self.driver.ensure_connectivity()

    def send(self, payload):
        return self.driver.send(payload)

    def receive(self):
        return self.driver.receive()


class LoRaDriverPycom:
    """
    LoRa driver for Pycom MicroPython
    """

    def __init__(self, network_manager, settings):
        self.network_manager = network_manager
        self.settings = settings

    def start(self):
        """ Start driver """

        from network import LoRa

        if self.settings.get('networking.lora.region')   == 'AS923':
            lora_region = LoRa.AS923
        elif self.settings.get('networking.lora.region') == 'AU915':
            lora_region = LoRa.AU915
        elif self.settings.get('networking.lora.region') == 'EU868':
            lora_region = LoRa.EU868
        elif self.settings.get('networking.lora.region') == 'US915':
            lora_region = LoRa.US915

        lora_adr = self.settings.get('networking.lora.adr') or False

        self.lora_socket = None

        self.lora = LoRa(mode=LoRa.LORAWAN, region=lora_region, adr=lora_adr)

        # Restore LoRa state from NVRAM after waking up from DEEPSLEEP.
        # Otherwise, reset LoRa NVRAM and rejoin.
        if platform_info.vendor in [platform_info.MICROPYTHON.Vanilla, platform_info.MICROPYTHON.Pycom]:
            import machine
            # Restore LoRaWAN status after wake up from deep sleep
            if machine.reset_cause() == machine.DEEPSLEEP_RESET:
                self.lora.nvram_restore()
                log.info('[LoRa] LoRaWAN state restored from NVRAM after deep sleep')
            # Otherwise reset LoRaWAN status and deep sleep interval from NVRAM
            else:
                self.lora.nvram_erase()
                log.info('[LoRa] LoRaWAN state erased from NVRAM. Rejoin forced')

                if platform_info.vendor == platform_info.MICROPYTHON.Pycom:
                    import pycom
                    nvram_get   = pycom.nvs_get
                    nvram_erase = pycom.nvs_erase
                elif platform_info.vendor == platform_info.MICROPYTHON.Vanilla:
                    import esp32
                    nvram_get   = esp32.nvs_get
                    nvram_erase = esp32.nvs_erase

                try:
                    if nvram_get('deepsleep') is not None:
                        nvram_erase('deepsleep')
                        log.info('[LoRa] Deep sleep interval erased from NVRAM. Return to settings value')
                except:
                    pass

        if not self.lora.has_joined():
            import binascii

            datarate_join = self.settings.get('networking.lora.datarate_join')

            # Over-the-Air Activation (OTAA)
            if self.settings.get('networking.lora.activation') == 'otaa':
                app_eui = binascii.unhexlify(self.settings.get('networking.lora.otaa.application_eui'))
                app_key = binascii.unhexlify(self.settings.get('networking.lora.otaa.application_key'))

                log.info('[LoRa] Attaching to the LoRaWAN network using OTAA')
                if self.settings.get('networking.lora.otaa.device_eui') is None:
                    self.lora.join(activation=LoRa.OTAA, auth=(app_eui, app_key), timeout=0, dr=datarate_join)
                else:
                    dev_eui = binascii.unhexlify(self.settings.get('networking.lora.otaa.device_eui'))
                    self.lora.join(activation=LoRa.OTAA, auth=(dev_eui, app_eui, app_key), timeout=0, dr=datarate_join)

            # Activation by Personalization (ABP)
            elif self.settings.get('networking.lora.activation') == 'abp':
                import struct
                dev_addr = struct.unpack(">l", binascii.unhexlify(self.settings.get('networking.lora.abp.device_address')))[0]
                nwk_swkey = binascii.unhexlify(self.settings.get('networking.lora.abp.network_session_key'))
                app_swkey = binascii.unhexlify(self.settings.get('networking.lora.abp.app_session_key'))

                log.info('[LoRa] Attaching to the LoRaWAN network using ABP')
                self.lora.join(activation=LoRa.ABP, auth=(dev_addr, nwk_swkey, app_swkey), timeout=0, dr=datarate_join)

    def ensure_connectivity(self):

        self.wait_for_join()

        if self.lora_joined:
            if self.lora_socket is None:
                try:
                    self.create_socket()
                except:
                    log.exception("[LoRa] Could not create LoRa socket")
        else:
            log.error("[LoRa] Could not join network")

    def wait_for_join(self):
        """ wait for device activation to complete """

        self.lora_joined = None
        while not self.lora.has_joined():
            log.info('[LoRa] Not joined yet...')
            time.sleep(self.settings.get('networking.lora.otaa.join_check_interval', 2.5))

        self.lora_joined = self.lora.has_joined()

        if self.lora_joined:
            log.info('[LoRa] Joined successfully')
        else:
            log.info('[LoRa] Failed to join network')

        return self.lora_joined

    def create_socket(self):
        """ Create socket for LoRa communication """

        # create a lora socket
        self.socket = socket.socket(socket.AF_LORA, socket.SOCK_RAW)

        self.socket.settimeout(6.0)

        self.lora_socket = True
        log.info('[LoRa] Socket created')

        return self.lora_socket

    def send(self, payload):
        """
        Send a LoRa packet.
        :param payload:
        """
        self.socket.setblocking(True)

        success = self.socket.send(payload)

        self.socket.setblocking(False)

        self.lora.nvram_save()

        return success

    def receive(self):
        """
        Receive a LoRa packet.
        """

        try:
            rx, port = self.socket.recvfrom(256)
        except socket.timeout:
            log.info('[LoRa] No packet received within receive window')

        if rx:
            log.info('[LoRa] Received: {} on port: {}'.format(rx, port))

        return rx, port


class LoRaDriverDragino:
    """
    LoRa driver for Dragino LoRa/GPS HAT on Raspberry Pi.
    """

    def __init__(self, network_manager, settings):
        self.network_manager = network_manager
        self.settings = settings

        self.dragino = None
        self.setup()

    def setup(self):
        log.info('[LoRa] Setting up LoRa')

        # Monkeypatch SX127x library
        from dragino.SX127x.board_config import BOARD
        BOARD.DIO3 = None

        from dragino.dragino import Dragino, LoRaWANAuthentication, LoRaWANConfig

        # Over-the-Air Activation (OTAA)
        if self.settings.get('networking.lora.activation') == 'otaa':
            lora_auth = LoRaWANAuthentication(auth_mode='OTAA',
                                              deveui=self.settings.get('networking.lora.otaa.device_eui'),
                                              appeui=self.settings.get('networking.lora.otaa.application_eui'),
                                              appkey=self.settings.get('networking.lora.otaa.application_key'))

        # Activation by Personalization (ABP)
        elif self.settings.get('networking.lora.activation') == 'abp':
            lora_auth = LoRaWANAuthentication(auth_mode='ABP',
                                              devaddr=self.settings.get('networking.lora.abp.device_address'),
                                              nwskey=self.settings.get('networking.lora.abp.network_session_key'),
                                              appskey=self.settings.get('networking.lora.abp.app_session_key'))

        lora_config = LoRaWANConfig(auth=lora_auth)
        self.dragino = Dragino(config=lora_config, logging_level=logging.DEBUG)

    def start(self):
        log.info('[LoRa] Starting join')
        self.dragino.join()

    def ensure_connectivity(self):
        while not self.dragino.registered():
            log.info('[LoRa] Waiting for connectivity')
            time.sleep(1)

    def send(self, payload):
        self.dragino.send_bytes(list(payload))
        return len(payload)

    def receive(self):
        return None, None



# -*- coding: utf-8 -*-
# (c) 2019 Richard Pobering <richard@hiveeyes.org>
# (c) 2019 Andreas Motl <andreas@hiveeyes.org>
# License: GNU General Public License, Version 3
from terkin import logging
from terkin.sensor import SensorManager, AbstractSensor

log = logging.getLogger(__name__)


def includeme(sensor_manager: SensorManager, sensor_info):
    """
    Create HX711 sensor object.

    :param sensor_manager:
    :param sensor_info:

    :return: sensor_object
    """
    sensor_object = HX711Sensor(settings=sensor_info)

    sensor_object.family = 'scale'
    sensor_object.set_address(sensor_info.get('number', sensor_info.get('address', 0)))
    sensor_object.register_pin('dout', sensor_info['pin_dout'])
    sensor_object.register_pin('pdsck', sensor_info['pin_pdsck'])
    sensor_object.register_parameter('scale', float(sensor_info['scale']))
    sensor_object.register_parameter('offset', float(sensor_info['offset']))
    sensor_object.register_parameter('gain', sensor_info.get('gain', 128))
    sensor_object.register_parameter('dualchannel', sensor_info.get('dualchannel', False))

    if "scaleB" in sensor_info:
        sensor_object.register_parameter('scaleB', float(sensor_info['scaleB']))
    if "offsetB" in sensor_info:
        sensor_object.register_parameter('offsetB', float(sensor_info['offsetB']))

    # Select driver module. Use "gerber" (vanilla) or "heisenberg" (extended).
    # hx711_sensor.select_driver('gerber')
    sensor_object.select_driver('heisenberg')

    return sensor_object


class HX711Sensor(AbstractSensor):
    """A generic HX711 sensor component wrapping possibly
    different hardware driver variants.
    
    After some boring parameter juggling, this sensors'
    ``read()`` method actually calls the hardware driver
    using ``self.loadcell.read_median()``.


    """

    def __init__(self, settings=None):

        super().__init__(settings=settings)

        # The driver class.
        self.driver_class = None

        # The driver instance.
        self.loadcell = None

    def select_driver(self, name='gerber'):
        """

        :param name:  (Default value = 'gerber')

        """

        # Use vanilla HX711 library by David Gerber.
        if name == 'gerber':
            from terkin.lib.hx711 import HX711

        # Use improved HX711 library by Ralf Lindlein.
        elif name == 'heisenberg':
            from terkin.lib.hx711_heisenberg import HX711Heisenberg as HX711

        # Error out for unknown hardware driver.
        else:
            raise ValueError('ERROR: Unknown HX711 hardware driver "{}"'.format(name))

        log.info('Selected HX711 hardware driver "{}"'.format(name))
        self.driver_class = HX711

    def start(self):
        """ """

        # Hardware parameters and configuration settings.
        pin_dout = self.pins['dout']
        pin_pdsck = self.pins['pdsck']
        gain = self.parameter.get('gain', 128)
        scale = self.parameter['scale']
        offset = self.parameter['offset']

        # Initialize the HX711 hardware driver.
        log.info('Initializing HX711 sensor with '
                 'pin_dout={}, pin_pdsck={}, gain={}, scale={}, offset={}'.format(pin_dout, pin_pdsck, gain, scale, offset))
        if self.parameter['dualchannel']:
            scaleB = self.parameter['scaleB']
            offsetB = self.parameter['offsetB']
            log.info('Initializing HX711 sensor channel B with gain=32 '
                    'scale={}, offset={}'.format(scaleB, offsetB))

        try:
            self.loadcell = self.driver_class(pin_dout, pin_pdsck, gain)

            # Configure the HX711 driver.
            if self.parameter['scale'] is not None:
                self.loadcell.set_scale(self.parameter['scale'])
            if self.parameter['offset'] is not None:
                self.loadcell.set_offset(self.parameter['offset'])

            return True

        except Exception as ex:
            log.exc(ex, 'HX711 hardware driver failed. Reason: {}'.format(ex))

    def read(self):
        """ """
        if self.loadcell is None:
            return self.SENSOR_NOT_INITIALIZED

        address = self.address

        #log.info('Acquire reading from HX711')
        readingA = self.loadcell.get_reading()

        # with dualchannel 3 data sets are generated instead of one
        # 1 - cumulated weight, no parameter, 2 - channel A reading with parameters, 3 - same for channel B
        if self.parameter['dualchannel']:   # read channel B
            self.loadcell.set_gain(32)      # switch to channel B
            self.loadcell.set_scale(self.parameter['scaleB'])
            self.loadcell.set_offset(self.parameter['offsetB'])
            readingB = self.loadcell.get_reading()
            self.loadcell.set_gain(self.parameter.get('gain', 128))      # switch everything back
            self.loadcell.set_scale(self.parameter['scale'])
            self.loadcell.set_offset(self.parameter['offset'])

            # Propagate main kg value.
            key = 'weight.{}'.format(address)
            totalweight = readingA.kg + readingB.kg
            effective_data = {
                key: totalweight
            }
            # Propagate _all_ values from HX711 channel A (raw, scale, offset, whatever).
            data = readingA.get_data()
            for key, value in data.items():
                effective_key = 'scaleA.{}.{}'.format(address, key)
                effective_data[effective_key] = value
            # Propagate _all_ values from HX711 channel B (raw, scale, offset, whatever).
            data = readingB.get_data()
            for key, value in data.items():
                effective_key = 'scaleB.{}.{}'.format(address, key)
                effective_data[effective_key] = value

        else:

            # Propagate main kg value.
            key = 'weight.{}'.format(address)
            effective_data = {
                key: readingA.kg
            }
            # Propagate _all_ values from HX711 (raw, scale, offset, whatever).
            data = readingA.get_data()
            for key, value in data.items():
                effective_key = 'scale.{}.{}'.format(address, key)
                effective_data[effective_key] = value

        return effective_data

    def power_on(self):
        """ """
        log.info('Powering up HX711')
        if self.loadcell:
            self.loadcell.power_up()

    def power_off(self):
        """ """
        log.info('Powering down HX711')
        if self.loadcell:
            self.loadcell.power_down()



# -*- coding: utf-8 -*-
# (c) 2019 Richard Pobering <richard@hiveeyes.org>
# (c) 2019 Andreas Motl <andreas@hiveeyes.org>
# License: GNU General Public License, Version 3
import logging
from logging import Logger, StreamHandler, Formatter
from logging import DEBUG, INFO, WARNING, ERROR, CRITICAL
from terkin.util import PycomChronometer, get_platform_info
from umal import GenericChronometer


# Keep track of time since boot.
_chrono = None


def get_chronometer():
    """
    Get a timer that counts from boot.
    """
    global _chrono
    if _chrono:
        return _chrono
    platform_info = get_platform_info()
    if platform_info.vendor == platform_info.MICROPYTHON.Pycom:
        _chrono = PycomChronometer()
    else:
        _chrono = GenericChronometer()
    return _chrono


if hasattr(logging, 'NullHandler'):
    NullHandler = logging.NullHandler
else:
    class NullHandler(logging.Handler):
        def handle(self, record):
            pass
    logging.NullHandler = NullHandler


class TimedLogRecord(logging.LogRecord):
    """ 
    Assign a time (since boot) to a log record.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        chrono = get_chronometer()
        try:
            self.tdelta = chrono.read()
        except:
            self.tdelta = 0


class ExtendedLogger(Logger):
    """ """

    def log(self, level, msg, *args):
        """

        :param level: 
        :param msg: 
        :param *args: 

        """
        from logging import _level

        if level >= (self.level or _level):
            record = TimedLogRecord(
                self.name, level, None, None, msg, args, None, None, None
            )

            if self.handlers:
                for hdlr in self.handlers:
                    hdlr.emit(record)


def getLogger(name=None, level=logging.INFO):
    """

    :param name:  (Default value = None)
    :param level:  (Default value = logging.INFO)

    """
    try:
        from logging import _loggers
    except ImportError:
        return logging.getLogger(name=name)

    if name is None:
        name = "root"
    if name in _loggers:
        return _loggers[name]
    else:
        l = ExtendedLogger(name)
        l.setLevel(level)

        sh = StreamHandler()

        #log_format = '[%(tdelta)s] %(levelname)s %(message)s'
        log_format = '%(tdelta)10.4f [%(name)-28s] %(levelname)-7s: %(message)s'

        sh.setFormatter(Formatter(log_format, style='%'))

        l.addHandler(sh)
    _loggers[name] = l
    return l


def noop(*args, **kwargs):
    """

    :param *args: 
    :param **kwargs: 

    """
    return


loggers_backup = {
    'Logger': Logger.log,
    'ExtendedLogger': ExtendedLogger.log,
}


def disable_logging():
    """ 
    Disabe logging.
    """
    Logger.log = noop
    ExtendedLogger.log = noop


def enable_logging():
    """ 
    Enable logging.
    """
    Logger.log = loggers_backup['Logger']
    ExtendedLogger.log = loggers_backup['ExtendedLogger']



# -*- coding: utf-8 -*-
# (c) 2016-2020 Clemens Gruber <clemens@hiveeyes.org>
# (c) 2017-2020 Andreas Motl <andreas@hiveeyes.org>
# License: GNU General Public License, Version 3
import sys
from terkin import logging
from terkin.sensor import SensorManager


log = logging.getLogger(__name__)


class ScaleAdjustment:
    """
    Python port of Open Hive scale adjustment sketch.

    https://github.com/hiveeyes/arduino/blob/master/scale-adjust/HX711/scale-adjust-hx711.ino
    """

    def __init__(self, sensor_manager: SensorManager):
        self.sensor_manager = sensor_manager

        # Sensor object of selected scale.
        self.sensor = None

        # Raw loadcell value when scale is unloaded.
        self.value_unloaded = None

        # Raw loadcell value when scale is loaded.
        self.value_loaded = None

        # Known weight put on the scale.
        self.adjustment_weight = None

    def start_wizard(self):
        self.cls()
        self.print_section('Weight scale adjustment', '=')
        self.select_sensor()
        self.start_sensor()
        self.read_unloaded()
        self.read_loaded()
        self.done()

    def select_sensor(self):
        while True:
            try:
                self.print_section('Step 1: Select sensor device')
                print()
                sensors = list(self.sensor_manager.get_sensors_by_family('scale'))
                for index, sensor in enumerate(sensors):
                    print('{}: {}'.format(index, sensor.settings.get('id')))
                print()
                sensor_index = int(input('Type index: '))
                self.sensor = sensors[sensor_index]
                print()
                break
            except KeyboardInterrupt:
                break
            except Exception as ex:
                log.exc(ex, 'Error while selecting sensor')

    def start_sensor(self):
        if hasattr(self.sensor, 'power_on'):
            self.sensor.power_on()
        if hasattr(self.sensor, 'loadcell') and hasattr(self.sensor.loadcell, 'initialize'):
            self.sensor.loadcell.initialize()

    def read_unloaded(self):
        while True:
            try:
                self.print_section('Step 2: Unloaded scale')
                print('Please remove all weight from the scale.')
                print('When done, press any key to continue.')
                self.wait_for_keypress()
                self.value_unloaded = float(self.sensor.loadcell.get_reading().raw)
                print()
                break
            except KeyboardInterrupt:
                break
            except Exception as ex:
                log.exc(ex, 'Error while reading unloaded scale')

    def read_loaded(self):
        while True:
            try:
                self.print_section('Step 3: Loaded scale')
                print('Please load the scale with a known weight.')
                print('When done, input the kilogram value.')
                print()

                self.adjustment_weight = float(input('Weight (kg): ').replace(',', '.'))
                self.value_loaded = float(self.sensor.loadcell.get_reading().raw)
                print()
                break
            except KeyboardInterrupt:
                break
            except Exception as ex:
                log.exc(ex, 'Error while reading loaded scale')

    def done(self):
        """
        Compute and output loadCellZeroOffset and loadCellKgDivider.
        """
        self.print_section('Step 4: Configure settings')
        print('Please use the computed values for configuring your settings.py')
        print()

        offset = self.value_unloaded
        divider = (self.value_loaded - self.value_unloaded) / self.adjustment_weight
        print('offset:', offset)
        print('scale: ', divider)

    def cls(self):
        sys.stdout.write(b"\x1b[2J")
        sys.stdout.write(b"\x1b[0;0H")

    def wait_for_keypress(self):
        sys.stdin.read(1)

    def print_section(self, message, char='-'):
        print()
        length = len(message)
        print(char * length)
        print(message)
        print(char * length)



# -*- coding: utf-8 -*-
# (c) 2019 Diren Senger <diren@uni-bremen.de>
# (c) 2019 Vincent Kuhlen <vkuhlen@uni-bremen.de>
# (c) 2019 Richard Pobering <richard@hiveeyes.org>
# (c) 2019 Andreas Motl <andreas@hiveeyes.org>
# License: GNU General Public License, Version 3
#
# Web server documentation:
# https://github.com/jczic/MicroWebSrv
#
# Original sources:
# https://github.com/hiveeyes/terkin-datalogger/pull/12/files
# https://github.com/Hiverize/FiPy/blob/master/webserver.py
#
import json

import machine

from copy import deepcopy
from MicroWebSrv2 import MicroWebSrv2, WebRoute, HttpRequest, GET, POST, PUT

from terkin import logging
from terkin.sensor.common import serialize_som

log = logging.getLogger(__name__)


# Global reference to webserver object.
webserver = None


class TerkinHttpApi:
    """ """

    device = None
    settings = None
    storage = None
    headers = None

    def __init__(self, device=None, settings=None, storage=None):

        TerkinHttpApi.device = device
        TerkinHttpApi.settings = settings
        TerkinHttpApi.storage = storage
        TerkinHttpApi.headers = {
            # Work around troubles with CORS in development.
            'Access-Control-Allow-Origin': '*',

            # Put application name and version into HTTP headers.
            'Application-Name': TerkinHttpApi.device.application_info.name,
            'Application-Version': TerkinHttpApi.device.application_info.version,
        }

        # Webserver settings.
        webserver_port = settings.get('services.api.http.port', 80)

        # Conditionally initialize webserver.
        # TCP port 80 and files in /flash/www.
        # TODO: Make port and htdocs folder configurable.
        global webserver
        if webserver is None:
            log.info('Creating new HTTP server object')
            webserver = MicroWebSrv2()
            webserver.BindAddress = ('0.0.0.0', webserver_port)
        self.webserver = webserver

    def start(self):
        """ """
        if self.webserver.IsRunning:
            log.info('HTTP server already started')
        else:
            log.info('Starting HTTP server')
            self.webserver.SetEmbeddedConfig()
            self.webserver.StartManaged()

    def stop(self):
        self.webserver.Stop()

    def captive(self):
        """
        Configure transparent captive portal.

        Intercept all requests and redirect them to the device.
        """

        # Intercept DNS inquiries.
        from microDNSSrv import MicroDNSSrv
        MicroDNSSrv.Create({'*': '192.168.4.1'})

        # All pages not found will be redirected here.
        self.webserver.NotFoundURL("http://hiverize.wifi")


    # ======
    # System
    # ======

    @WebRoute(GET, '/status')
    @WebRoute(POST, '/status')
    def status(microWebSrv2, request: HttpRequest):
        TerkinHttpApi.respond_text(request, 'OK')

    @WebRoute(GET, '/about')
    def about(microWebSrv2, request: HttpRequest):
        TerkinHttpApi.respond_text(request, TerkinHttpApi.device.application_info.fullname)

    @WebRoute(POST, '/restart')
    def restart(microWebSrv2: MicroWebSrv2, request: HttpRequest):

        TerkinHttpApi.respond_text(request, 'ACK')

        def do_reset():
            """ """
            log.info('Resetting device')
            microWebSrv2.Stop()
            machine.reset()

        try:
            import _thread
            _thread.start_new_thread(do_reset, ())
        except:
            do_reset()

    # ===========
    # Application
    # ===========

    @WebRoute(GET, '/api/v1/settings')
    def get_settings(microWebSrv2, request: HttpRequest):
        try:
            query_params = request.QueryParams
            format = query_params.get('format', 'json')
            if format == 'json':
                headers = dict(TerkinHttpApi.headers)
                headers.update({'Content-Disposition': 'attachment; filename="settings.json"'})
                TerkinHttpApi.respond_json(request, TerkinHttpApi.settings.to_dict())

            elif format == 'python':
                request.Response._headers.update(TerkinHttpApi.headers)
                request.Response.ReturnFile(filename='/flash/settings.py', attachmentName='settings.py')

            else:
                request.Response.ReturnNotFound()

        except Exception as ex:
            log.exc(ex, 'GET settings request failed')
            request.Response.ReturnBadRequest()

    @WebRoute(PUT, '/api/v1/settings')
    def put_settings(microWebSrv2, request: HttpRequest):
        try:

            # Sanity checks.
            # TODO: Validate the request.
            content_type = request.ContentType

            if content_type is None:
                request.Response.ReturnNotFound()

            if content_type.startswith('application/json'):
                data = request.GetPostedJSONObject()
                TerkinHttpApi.settings.save('settings-user.json', json.dumps(data))
                TerkinHttpApi.respond_text(request, 'ACK')

            elif content_type.startswith('text/plain') or content_type.startswith('application/octet-stream'):
                """
                buffer = TerkinHttpApi.read_request(httpClient)
                #print('body:')
                #print(body)
                TerkinHttpApi.settings.save('settings.py', buffer)
                TerkinHttpApi.respond_text(httpResponse, 'ACK')
                """
                request.Response.ReturnBadRequest()

            else:
                request.Response.ReturnNotFound()

        except Exception as ex:
            log.exc(ex, 'PUT settings request failed')
            request.Response.ReturnInternalServerError()

    @WebRoute(GET, '/api/v1/setting')
    def get_setting(microWebSrv2, request: HttpRequest):
        try:
            query_data = request.QueryParams
            name = query_data['name']
            log.info('Getting configuration setting "{}"'.format(name))
            value = TerkinHttpApi.settings.get(name)
            log.info('Configuration setting "{}" is "{}"'.format(name, value))
            TerkinHttpApi.respond_json(request, value)

        except Exception as ex:
            log.exc(ex, 'GET setting request failed')
            request.Response.ReturnInternalServerError()

    @WebRoute(PUT, '/api/v1/setting')
    def put_setting(microWebSrv2, request: HttpRequest):
        try:
            query_data = request.QueryParams
            name = query_data['name']
            value = request.GetPostedJSONObject()
            log.info('Setting configuration setting "{}" to "{}"'.format(name, value))
            TerkinHttpApi.settings[name] = value

            value = TerkinHttpApi.settings.get(name)
            log.info('Re-reading configuration setting "{}" as "{}"'.format(name, value))
            TerkinHttpApi.respond_json(request, value)

        except Exception as ex:
            log.exc(ex, 'PUT setting request failed')
            request.Response.ReturnInternalServerError()

    @WebRoute(GET, '/api/v1/reading/last')
    def get_settings(microWebSrv2, request: HttpRequest):
        try:
            TerkinHttpApi.respond_json(request, TerkinHttpApi.storage.last_reading)
            return

        except Exception as ex:
            log.exc(ex, 'GET last reading request failed')
            request.Response.ReturnInternalServerError()

        request.Response.ReturnNotFound()

    def read_request(request: HttpRequest):
        # Observations show request payloads are capped at ~4308 bytes.
        # https://github.com/jczic/MicroWebSrv/issues/51
        from uio import StringIO
        buffer = StringIO()
        while True:
            try:
                log.info('Reading 4000 bytes from network')
                payload = request.Content
                if not payload:
                    log.info('Reading finished')
                    raise StopIteration()
                log.info('Writing {} bytes to buffer'.format(len(payload)))
                buffer.write(payload)
            except:
                break
        log.info('Rewinding buffer')
        buffer.seek(0)
        return buffer

    @WebRoute(GET, '/api/v1/peripherals/buses')
    def sensor_index(microWebSrv2, request: HttpRequest):
        sensor_manager = TerkinHttpApi.device.application_info.application.sensor_manager
        som_info = serialize_som(sensor_manager.buses)
        TerkinHttpApi.respond_json(request, som_info)

    @WebRoute(GET, '/api/v1/peripherals/sensors')
    def sensor_index(microWebSrv2, request: HttpRequest):
        sensor_manager = TerkinHttpApi.device.application_info.application.sensor_manager
        som_info = serialize_som(sensor_manager.sensors)
        TerkinHttpApi.respond_json(request, som_info)

    @WebRoute(GET, '/api/v1/sensors/ds18b20')
    def sensor_index_ds18b20(microWebSrv2, request: HttpRequest):
        sensor_manager = TerkinHttpApi.device.application_info.application.sensor_manager
        ds18b20_sensors = []
        #print('sensor_manager.buses:', sensor_manager.buses)

        # Collect information about all DS18B20 sensors connected to all 1-Wire buses.
        for sensor in sensor_manager.sensors:
            if not hasattr(sensor, 'type') or sensor.type != 'DS18B20':
                continue
            bus_info = {
                'bus': sensor.bus.name,
                'pin': sensor.bus.pins.get('data'),
            }
            for address in sensor.bus.get_devices_ascii():
                sensor_info = deepcopy(bus_info)
                sensor_info['address'] = address
                sensor_info['description'] = sensor.get_device_description(address)
                ds18b20_sensors.append(sensor_info)

        TerkinHttpApi.respond_json(request, ds18b20_sensors)

    @WebRoute(GET, '/api/v1/sensor/<sensor>')
    def sensor_info(microWebSrv2, request: HttpRequest, routeArgs):
        sensor = routeArgs['sensor']

    # ====
    # Demo
    # ====

    @WebRoute(GET, '/echo/<slot>')
    @WebRoute(POST, '/echo/<slot>')
    def echo(microWebSrv2, request: HttpRequest, routeArgs):

        # Collect information from HTTP request.
        content_type = request.ContentType
        query_data = request.QueryParams

        if content_type.startswith('application/json'):
            data = request.GetPostedJSONObject()

        elif content_type.startswith('application/x-www-form-urlencoded'):
            data = request.GetPostedURLEncodedForm()

        else:
            data = None

        # Bundle request information.
        payload = {
            'path': routeArgs,
            'content_type': content_type,
            'query': query_data,
            'data': data,
        }
        TerkinHttpApi.respond_json(request, payload)

    # =========
    # Utilities
    # =========

    def respond_text(request, text):
        request.Response._headers.update(TerkinHttpApi.headers)
        request.Response._contentType = 'text/plain'
        request.Response._contentCharset = 'utf-8'
        request.Response.ReturnOk(content=text)

    def respond_json(request, data):
        request.Response._headers.update(TerkinHttpApi.headers)
        request.Response.ReturnOkJSON(obj=data)



import array

mandatory_fields = [
    'latitude', 'longitude', 'roll', 'pitch', 'speed', 'cog',
    'temperature', 'pressure', 'humidity', 'battery_voltage',
]


def create_payload(data):
    #print('DATA:', data)

    # Sanity checks
    for field in mandatory_fields:
        if field not in data or data[field] is None:
            raise KeyError('Mandatory field "{}" required'.format(field))

    lat, lon = convert_latlon(data.get('latitude'), data.get('longitude'))
    roll, pitch, bat = convert_pytrack_sensors(data.get('roll'), data.get('pitch'), data.get('battery_voltage'))
    speedkmh, cog = convert_speed(data.get('speed'),data.get('cog'))
    temp, pressure, humidity = convert_bme280(data.get('temperature'), data.get('pressure'), data.get('humidity'))

    # payload = array.array('B', [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0])
    payload = array.array('B', [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0])
    payload[0] = lat
    payload[1] = (lat >> 8)
    payload[2] = (lat >> 16)

    payload[3] = lon
    payload[4] = (lon >> 8)
    payload[5] = (lon >> 16)

    payload[6] = roll
    payload[7] = (roll >> 8)
    payload[8] = (roll >> 16)

    payload[9] = pitch
    payload[10] = (pitch >> 8)
    payload[11] = (pitch >> 16)

    payload[12] = bat
    payload[13] = (bat >> 8)
    payload[14] = (bat >> 16)

    payload[15] = cog
    payload[16] = (cog >> 8)
    payload[17] = (cog >> 16)

    payload[18] = speedkmh
    payload[19] = (speedkmh >> 8)
    payload[20] = (speedkmh >> 16)

    payload[21] = temp
    payload[22] = (temp >> 8)
    payload[23] = (temp >> 16)

    payload[24] = pressure
    payload[25] = (pressure >> 8)
    payload[26] = (pressure >> 16)

    return payload


def convert_pytrack_sensors(r, p, b):
    r = int((r + 360) * 10000)
    p = int((p + 360) * 10000)
    b = int(b * 10000)
    return r, p, b


def convert_speed(s, c):
    s = int((float(s) + 999) * 1000)
    c = int((float(c) + 999) * 1000)
    return s, c


def convert_bme280(t, p, h):
    t = int((float(t) + 999) * 1000)
    p = int((float(p) + 999) * 1000)
    h = int((float(h) + 999) * 1000)
    return t, p, h


def convert_latlon(lat, lon):
    lat = int((float(lat) + 90)*10000)
    lon = int((float(lon) + 180)*10000)
    # alt = int((altitude) * 10)
    # hdop = int((hdop) * 10)
    return lat, lon



# -*- coding: utf-8 -*-
# (c) 2019 Richard Pobering <richard@hiveeyes.org>
# (c) 2019 Andreas Motl <andreas@hiveeyes.org>
# License: GNU General Public License, Version 3
from terkin import logging
from terkin.sensor import SensorManager, AbstractSensor
from terkin.util import get_platform_info

log = logging.getLogger(__name__)
platform_info = get_platform_info()


def includeme(sensor_manager: SensorManager, sensor_info):
    """
    Create SI7021 sensor object.

    :param sensor_manager:
    :param sensor_info:

    :return: sensor_object
    """
    sensor_object = SI7021Sensor(settings=sensor_info)
    sensor_bus = sensor_manager.get_bus_by_name(sensor_info.get('bus'))
    sensor_object.acquire_bus(sensor_bus)

    return sensor_object


class SI7021Sensor(AbstractSensor):
    """
    A generic SI7021 sensor component.
    Used by terkin/datalogger to register and read() from this sensor.
    start() & read() are mandatory.
    """

    def __init__(self, settings=None):

        super().__init__(settings=settings)

        # Can be overwritten by ``.set_address()``.
        self.address = self.settings.get('address', 0x40)

    def start(self):
        """Getting the bus"""
        if self.bus is None:
            raise KeyError("Bus missing for SI7021Sensor")

        # Initialize the hardware driver.
        try:

            # Vanilla MicroPython 1.11 and Pycom MicroPython 1.9.4
            if platform_info.vendor in [platform_info.MICROPYTHON.Vanilla, platform_info.MICROPYTHON.Pycom]:
                from SI7021 import SI7021
                self.driver = SI7021(i2c=self.bus.adapter)

            # Adafruit CircuitPython
            elif platform_info.vendor == platform_info.MICROPYTHON.RaspberryPi:
                import adafruit_si7021
                self.driver = adafruit_si7021.SI7021(i2c_bus=self.bus.adapter, address=self.address)

            else:
                raise NotImplementedError('Si7021 driver not implemented on this platform')

            return True

        except Exception as ex:
            log.exc(ex, 'SI7021 hardware driver failed')

    def read(self):
        """ """

        if self.bus is None or self.driver is None:
            return self.SENSOR_NOT_INITIALIZED

        #log.info('Acquire reading from SI7021')

        data = {}

        # Vanilla MicroPython 1.11 and Pycom MicroPython 1.9.4
        if platform_info.vendor in [platform_info.MICROPYTHON.Vanilla, platform_info.MICROPYTHON.Pycom]:

            # Prepare readings.
            values = {
                "temperature": self.driver.temperature(),
                "humidity": self.driver.humidity(),
            }

        # Adafruit CircuitPython
        elif platform_info.vendor == platform_info.MICROPYTHON.RaspberryPi:

            values = {
                "temperature": self.driver.temperature,
                "humidity": self.driver.relative_humidity,
            }

        # Build telemetry payload.
        fieldnames = values.keys()
        for name in fieldnames:
            fieldname = self.format_fieldname(name, hex(self.address))
            value = values[name]
            data[fieldname] = value

        if not data:
            log.warning("I2C device {} has no value: {}".format(self.address, data))

        log.debug("I2C data:     {}".format(data))

        return data




# -*- coding: utf-8 -*-
# (c) 2019 Jan Hoffmann <jan.hoffmann@bergamsee.de>
# (c) 2019 Andreas Motl <andreas@hiveeyes.org>
# License: GNU General Public License, Version 3
from terkin import logging
from terkin.model import SensorReading
from terkin.sensor import SensorManager, AbstractSensor

log = logging.getLogger(__name__)


def includeme(sensor_manager: SensorManager, sensor_info):
    """
    Create BME280 sensor object.

    :param sensor_manager:
    :param sensor_info:

    :return: sensor_object
    """
    sensor_object = PytrackSensor(settings=sensor_info)
    sensor_bus = sensor_manager.get_bus_by_name(sensor_info.get('bus'))
    sensor_object.acquire_bus(sensor_bus)

    return sensor_object


class PytrackSensor(AbstractSensor):
    """A Pytrack sensor component."""

    def __init__(self, settings=None):

        super().__init__(settings=settings)

        # The driver instance.
        self.readings = None
        self.bus = None

        self.sensor = None
        self.l76 = None
        self.lis2hh12 = None

    def start(self):
        """Getting the bus"""
        if self.bus is None:
            raise KeyError("I2C bus missing")

        # Initialize the hardware driver.
        try:
            from pytrack import Pytrack
            self.sensor = Pytrack(i2c=self.bus.adapter)
        except Exception as ex:
            log.exc(ex, 'Pytrack hardware driver failed')
            raise

        # Initialize the L76GNS sensor driver.
        try:
            from L76GNSV4 import L76GNSS
            self.l76 = L76GNSS(pytrack=self.sensor, timeout=5)
        except Exception as ex:
            log.exc(ex, 'Pytrack L76GNSS hardware driver failed')
            raise

        # Initialize the LIS2HH12 sensor driver.
        try:
            from LIS2HH12 import LIS2HH12
            self.lis2hh12 = LIS2HH12(pysense=self.sensor)
        except Exception as ex:
            log.exc(ex, 'Pytrack LIS2HH12 hardware driver failed')
            raise

    def read(self):
        """ """
        data = {'battery_voltage': float(self.sensor.read_battery_voltage())}
        # log.info('Acquire reading from Pytrack')

        # TODO: Add more data here.
        l76_data = self.read_l76gns()
        data.update(l76_data)
        lis2hh12_data = self.read_lis2hh12()
        data.update(lis2hh12_data)

        log.info("Pytrack data: {}".format(data))

        reading = SensorReading()
        reading.sensor = self
        reading.data = data

        return reading

    def read_lis2hh12(self):
        """ """

        data = {}

        try:
            data['roll'] = float(self.lis2hh12.roll())
            data['pitch'] = float(self.lis2hh12.pitch())
        except Exception as ex:
            log.exc(ex, 'No Roll/Pitch Data:')

        try:
            # Returns (self.x[0] * _mult, self.y[0] * _mult, self.z[0] * _mult)
            acceleration = self.lis2hh12.acceleration()
            data['acceleration_x'] = float(acceleration[0])
            data['acceleration_y'] = float(acceleration[1])
            data['acceleration_z'] = float(acceleration[2])
        except Exception as ex:
            log.exc(ex, 'No Acceleration Data')

        # TODO: Enable enable_activity_interrupt
        # try:
        #    data['activity'] = self.lis2hh12.activity()
        # except Exception as ex:
        #    log.exc(ex, 'No Acitivity Data')

        # TODO: Enable enable_activity_interrupt
        # try:
        #    self.lis2hh12.enable_activity_interrupt(self.settings['threshold'], self.settings['duration'])
        # except Exception as e:
        #    log.info('No enable_activity_interrupt Data:', e)

        return data

    def read_l76gns(self):
        """ """

        data = {}

        # returns the info about sattelites in view at this moment
        # even without the gps being fixed
        #try:
            #print(self.l76.gps_message('GSV', debug=False))
            # {'Elevation2': '', 'Elevation3': '', 'SNR2': '', 'SNR3': '', 'Elevation1': '10', 'SNR1': '',
            # 'Elevation4': '', 'Azimuth3': '', 'Azimuth2': '', 'Azimuth1': '086', 'SatelliteID3': '',
            # 'SatelliteID2': '', 'SatelliteID1': '82', 'SatellitesInView': '09', 'NofMessage': '3', 'Azimuth4': '',
            # 'SequenceNr': '3', 'NMEA': 'GLGSV', 'SatelliteID4': '', 'SNR4': ''}
        #except Exception as ex:
        #    log.exc(ex, 'No gps_messages')

        try:
            data['satellites_count'] = self.l76.gps_message('GGA', debug=False)['NumberOfSV']
        except Exception as ex:
            log.exc(ex, "No satellites Data")

        # Call this to start the machinery and actually get a fix.
        try:
            self.l76.get_fix()
        except Exception as ex:
            log.exc(ex, "No Fix Data")
        # Only read values when having a fix.
        if not self.l76.fixed():
            return data

        try:
            self.l76.coordinates()
        except Exception as ex:
            log.exc(ex, "No coordinates")
            # log.exc(ex, "Could not read coordinates")
            # raise

        # Read speed and orientation.
        try:
            speed = self.l76.get_speed()
            data['speed'] = float(speed.get('speed'))
            data['cog'] = float(speed.get('COG'))
        except Exception as ex:
            log.exc(ex, "Could not read Speed. Error")

        # Read position.
        try:
            # location = self.l76.get_location(MSL=True)
            location = self.l76.get_location()
            try:
                data['longitude'] = float(location.get('longitude'))
                data['latitude'] = float(location.get('latitude'))
                data['altitude'] = float(location.get('altitude'))
            except Exception as ex:
                log.exc(ex, 'No GPS Data')
        except Exception as ex:
            log.exc(ex, "Could not read location from L76 GNSS")

        try:
            utctime = self.l76.getUTCTime()
            if utctime:
                data['UTCTime'] = utctime
            utcdatetime = self.l76.getUTCDateTime()
            if utcdatetime:
                data['UTCDateTime'] = utcdatetime
            # TODO Needed?
            utcdatetimetuple = self.l76.getUTCDateTimeTuple()
            # if utcdatetimetuple:
            #    data['UTCDateTimeTuple'] = utcdatetimetuple
            log.info('DateTimeTuple: ' + str(utcdatetimetuple))
        except Exception as ex:
            log.exc(ex, 'No Time Data')

        # TODO: Sort into Sensor Metadata?
        #try:
            #data['chip_version'] = str(self.l76.get_chip_version())
            # {'ChipVersionID': 'L76LNR01A03S', 'time': '17:40', 'date': '2016/05/16',
            # 'command': 'R', 'PMTK': 'PERNO'}
            #data['dt_release'] = str(self.l76.get_dt_release())
            # {'ProductModel': 'Quectel-L76', 'ReleaseString''AXN_3.8_3333_16042918',
            # 'PMTK': 'PMTK705', 'BuildID': '0006', 'SDK': '1.0'}
        #except Exception as ex:
        #    log.exc(ex, 'No Chip Data')

        return data



# -*- coding: utf-8 -*-
# (c) 2019 Richard Pobering <richard@hiveeyes.org>
# (c) 2019 Andreas Motl <andreas@hiveeyes.org>
# License: GNU General Public License, Version 3
import os
import json
import types
try:
    import os_path
except ImportError:
    import os.path as os_path
from copy import deepcopy
from dotty_dict import dotty
from shutil import copyfileobj
from terkin import logging
from terkin.backup import backup_file
from terkin.util import ensure_directory, get_platform_info, deepupdate

log = logging.getLogger(__name__)


class TerkinConfiguration:
    """A flexible configuration manager."""

    USER_SETTINGS_FILE = 'settings-user.json'

    # Strip some settings when displaying configuration values
    # to prevent leaking sensible information into log files.
    # Todo: Should more things be optionally added here? E.g.
    # - Settings: ssid, endpoint, address
    # - Log: Device id, Networks (available|candidates), mac, ifconfig, Telemetry channel URI, MQTT broker
    protected_settings = [

        # WiFi
        'password',

        # LoRa
        'application_key', 'application_eui',
        
        # BEEP-BOB telemetry
        'key',
    ]

    def __init__(self):

        self.store = dotty()
        self.overlay = dotty()

        self.record = True

        self.compute_paths()

        log.info('Starting TerkinConfiguration on path "{}"'.format(self.CONFIG_PATH))
        #os.stat(self.CONFIG_PATH)

        if self.get('main.backup.enabled', False):
            try:
                log.info('Ensuring existence of backup directory at "{}"'.format(self.BACKUP_PATH))
                ensure_directory(self.BACKUP_PATH)
            except Exception as ex:
                log.exc(ex, 'Ensuring existence of backup directory at "{}" failed'.format(self.BACKUP_PATH))

    def __getitem__(self, key, default=None):
        return self.get(key, default=default)

    def __setitem__(self, key, value):
        return self.set(key, value)

    def __delitem__(self, key):
        del self.store[key]

    def compute_paths(self):
        platform_info = get_platform_info()
        if platform_info.vendor == platform_info.MICROPYTHON.Pycom:
            self.CONFIG_PATH = '/flash'
            self.BACKUP_PATH = '/flash/backup'
        elif platform_info.vendor == platform_info.MICROPYTHON.Vanilla:
            self.CONFIG_PATH = '/'
            self.BACKUP_PATH = '/backup'
        else:
            self.CONFIG_PATH = os.path.abspath('.')
            self.BACKUP_PATH = os.path.join(self.CONFIG_PATH, 'backup')

    def get(self, key, default=None):
        """

        :param key: 
        :param default:  (Default value = None)

        """
        return self.store.get(key, default=default)

    def set(self, key, value):
        """

        :param key: 
        :param value: 

        """
        self.store[key] = value
        if self.record:
            self.overlay[key] = value
            self.save(self.USER_SETTINGS_FILE, json.dumps(self.overlay.to_dict()))
        return value

    def setdefault(self, key, default=None):
        """

        :param key: 
        :param default:  (Default value = None)

        """
        # TODO: Add recording functionality.
        return self.store.setdefault(key, default=default)

    def add(self, data):
        """

        :param data: 

        """
        self.record = False
        try:
            if isinstance(data, types.ModuleType):
                blacklist = ['__name__', '__file__', '__class__']
                for key in dir(data):
                    if key in blacklist: continue
                    value = getattr(data, key)
                    self.store[key] = value
        except Exception as ex:
            log.exc(ex, 'Reading configuration settings failed')
        self.record = True

    def dump(self):
        """ """
        log.info('Configuration settings:')
        for key, value in self.store.items():
            if key.startswith('__'):
                continue
            thing = deepcopy(value)
            self.purge_protected_settings(thing)
            log.info('Section "{}": {}'.format(key, json.dumps(thing)))

    def purge_protected_settings(self, thing):
        """
        This function purges all protected pieces from the data structure
        holding the configuration settings in order to protect information
        leakage os sensitive information to STDOUT.

        :param thing: 
        """
        if isinstance(thing, dict):
            for key, value in thing.items():

                if key in self.protected_settings:
                    value = '## redacted ##'
                    thing[key] = value

                self.purge_protected_settings(value)

        elif isinstance(thing, list):
            for item in thing:
                self.purge_protected_settings(item)

    def to_dict(self):
        """ """
        return dict(self.store.to_dict())

    def add_user_file(self):
        """ """
        data = self.load(self.USER_SETTINGS_FILE)
        log.info('User settings: %s', data)
        if data is not None:
            self.overlay.update(data)
            deepupdate(self.store, self.overlay)

    def load(self, filename):
        """Load configuration file.

        :param filename: 

        """
        # Protect against directory traversals.
        filename = os_path.basename(filename)

        # Absolute path to configuration file.
        filepath = os_path.join(self.CONFIG_PATH, filename)

        # FIXME: Use ``os_path.exists``.
        try:
            os.stat(filepath)
        except OSError:
            return

        log.info('Reading configuration file {}'.format(filepath))
        try:
            with open(filepath, "r") as instream:
                payload = instream.read()
                data = json.loads(payload)
                return data

        except Exception as ex:
            log.exc(ex, 'Reading configuration from "{}" failed'.format(filepath))

    def save(self, filename, instream):
        """Save configuration file, with rotating backup.

        :param filename: 
        :param instream: 

        """
        import uos

        # Protect against directory traversals.
        filename = os_path.basename(filename)

        # Only allow specific filenames.
        if 'settings' not in filename:
            raise ValueError('Writing arbitrary files to the system is prohibited')

        # Absolute path to configuration file.
        filepath = os_path.join(self.CONFIG_PATH, filename)

        # Number of backup files to keep around.
        backup_count = self.get('main.backup.file_count', 7)

        # Backup configuration file.
        log.info('Backing up file {} to {}, keeping a history worth of {} files'.format(filepath, self.BACKUP_PATH, backup_count))
        backup_file(filepath, self.BACKUP_PATH, backup_count)

        # Overwrite configuration file.
        log.info('Saving configuration file {}'.format(filepath))
        with open(filepath, "w") as outstream:
            if isinstance(instream, str):
                outstream.write(instream)
            else:
                copyfileobj(instream, outstream)
            outstream.flush()

        uos.sync()



# -*- coding: utf-8 -*-
# (c) 2019 Richard Pobering <richard@hiveeyes.org>
# (c) 2019 Andreas Motl <andreas@hiveeyes.org>
# License: GNU General Public License, Version 3
import time
from terkin import logging

"""
==================
ESP32 Touch Sensor
==================

A touch-sensor system is built on a substrate which carries electrodes and 
relevant connections under a protective flat surface. When a user touches 
the surface, the capacitance variation is triggered and a binary signal is 
generated to indicate whether the touch is valid.

ESP32 can provide up to 10 capacitive touch pads / GPIOs. The sensing pads 
can be arranged in different combinations (e.g. matrix, slider), so that a 
larger area or more points can be detected. The touch pad sensing process 
is under the control of a hardware-implemented finite-state machine (FSM) 
which is initiated by software or a dedicated hardware timer.

- https://docs.pycom.io/tutorials/all/touch.html
- https://github.com/espressif/esp-iot-solution/blob/master/documents/touch_pad_solution/touch_sensor_design_en.md
- https://docs.espressif.com/projects/esp-idf/en/latest/api-reference/peripherals/touch_pad.html
- https://github.com/espressif/esp-idf/tree/58df1d93b/examples/peripherals/touch_pad_read
- https://www.instructables.com/id/ESP32-With-Capacitive-Touch-Button/
- https://microcontrollerslab.com/esp32-touch-sensor-button-example/

FiPy Pinout
-----------
- https://pycom.io/wp-content/uploads/2018/08/fipyPinoutNewCompN.pdf
"""

log = logging.getLogger(__name__)
#log.setLevel(logging.DEBUG)


class TouchPad:
    """From https://docs.pycom.io/tutorials/all/touch.html"""

    def __init__(self, pin, name, sensitivity, duration):
        self.name = name
        self.pin = pin
        self.sensitivity = sensitivity
        self.duration = duration

        self.touch = None
        self.last_press = 0
        self.pressed = False

        self.setup()

    def setup(self):
        """ """
        from machine import Touch
        self.touch = Touch(self.pin)
        self.touch.init_value(self.sensitivity)

    def read(self):
        """ """
        value = self.touch.read()
        log.debug('value: %s %s', self.name, value)
        return value

    def is_pressed(self):
        """ """
        # Todo: Review these magic values ``* 2 / 3``.
        threshold = self.touch.init_value() * 2 / 3
        if self.read() < threshold:
            self.pressed = True
        else:
            self.pressed = False
        return self.pressed

    def just_pressed(self):
        """ """
        now = time.ticks_ms()
        if now - self.last_press < self.duration:
            return True
        else:
            return False

    def set_press(self):
        """ """
        self.last_press = time.ticks_ms()



# -*- coding: utf-8 -*-
# (c) 2019-2020 Matthias Mehldau <wetter@hiveeyes.org>
# (c) 2019-2020 Andreas Motl <andreas@hiveeyes.org>
# License: GNU General Public License, Version 3
import ure
import time

from terkin import logging

log = logging.getLogger(__name__)


class SequansLTE:
    """
    Synopsis::

        sq = SequansLTE()
        sq.info()
        sq.firmware_info()

        sq.at('showphy')

    See also:
    - https://git.cicer.de/autonome-zelle/fipy-nbiot-rtd/blob/master/main.py

    """

    def __init__(self, network_manager, settings):
        self.network_manager = network_manager
        self.settings = settings

        from network import LTE
        self.lte = LTE()

        import machine
        self.chrono = machine.Timer.Chrono()
        self.chrono.start()

    def start(self):
        self.lte.init()
        self.attach()
        self.connect()

    def stop(self):
        self.lte.disconnect()
        time.sleep(0.25)

        self.lte.deinit()
        time.sleep(0.25)

    def attach(self):
        log.info('Attaching to LTE')
        self.lte.attach(band=self.settings.get('networking.lte.band'), apn=self.settings.get('networking.lte.apn'))

        self.chrono.reset()
        while True:

            log.info('Signal strength: {}'.format(self.get_signal_strength()))

            if self.lte.isattached():
                break

            if self.chrono.read() > self.settings.get('networking.lte.attach_timeout'):
                raise Exception('Attaching to LTE timed out')

            time.sleep(0.25)

    def connect(self):
        log.info('Connecting to LTE')
        self.lte.connect()

        self.chrono.reset()
        while True:

            if self.lte.isconnected():
                break

            if self.chrono.read() > self.settings.get('networking.lte.connect_timeout'):
                raise Exception('Connecting to LTE timed out')

            time.sleep(0.25)

    def imei(self):
        """
        Return IMEI.
        """
        return self.at('AT+CGSN=1')

    def info(self):
        """
        Get infos from Modem.
        """

        log.info('Signal strength: {}'.format(self.get_signal_strength()))

        self.at('RRC:setDbgPerm full')
        self.at('RRC:showcaps')
        self.at('showver')

        # https://forum.pycom.io/topic/4022/unable-to-update-gpy-modem-firmware/8
        #self.at('AT')
        #self.at('ATI')
        #self.at('ATZ')

    def get_signal_strength(self):
        csq_at = self.lte.send_at_cmd("AT+CSQ")
        csq_line_regex = ure.compile("\n")
        csq_line = csq_line_regex.split(csq_at)
        csq_string_regex = ure.compile(" ")
        csq_string = csq_string_regex.split(csq_line[1])
        csq_comma = csq_string[1]
        csq_num_regex = ure.compile(",")
        csq_num = csq_num_regex.split(csq_comma)
        csq = csq_num[0]
        return csq

    def at(self, command):
        """

        :param command:

        """
        return self.raw('AT!="{}"'.format(command))

    def raw(self, command):
        """

        :param command:

        """
        log.info('Sending: {}'.format(command))
        answer = self.lte.send_at_cmd(command)
        log.info('Answer:  {}'.format(answer))
        return answer

    def firmware_info(self):
        """ """
        import sqnsupgrade
        sqnsupgrade.info(verbose=True, debug=True)

    def unbrick(self):
        """ """
        raise NotImplementedError('https://forum.pycom.io/topic/4022/unable-to-update-gpy-modem-firmware/21')



# -*- coding: utf-8 -*-
# (c) 2019 Richard Pobering <richard@hiveeyes.org>
# (c) 2019 Andreas Motl <andreas@hiveeyes.org>
# License: GNU General Public License, Version 3
import ubinascii
from struct import pack, unpack

class BluetoothUuid:
    """

    *****
    About
    *****
    Work around the 128-bit UUID byte swap problem in Android's Bluetooth stack "Bluedroid".

    - https://forum.pycom.io/topic/5335/ble-decoding-service-uuid
    - https://community.hiveeyes.org/t/ble-gatt-auf-esp32-mit-micropython/2371/23


    ************
    Introduction
    ************

    UUID RFC-4122
    =============

    UUIDs are documented in RFC 4122. Here is an extract from that RFC:

    In the absence of explicit application or presentation protocol
    specification to the contrary, a UUID is encoded as a 128-bit object,
    as follows:

    The fields are encoded as 16 octets, with the sizes and order of the
    fields defined above, and with each field encoded with the Most
    Significant Byte first (known as network byte order aka big-endian).
    Note that the field names, particularly for multiplexed fields,
    follow historical practice.

    -- http://howtowriteaprogram.blogspot.com/2009/03/uuid-and-byte-order.html


    Bluetooth
    =========

    However, it looks like the "Supplement to Bluetooth Core Specification page 9 of 37" says:

    > All numerical multi-byte entities and values associated with the
      following data types shall use little-endian byte order.

    -- https://github.com/google/eddystone/issues/138#issuecomment-212332184


    ********
    And now?
    ********

    Testimonials
    ============

    While the "standard" UUID representation is little endian,
    the Android Bluetooth stack "Bluedroid" apparently isn't
    converting from little endian appropriately.
    -- https://stackoverflow.com/questions/35339982/strange-uuid-reversal-from-fetchuuidswithsdp
    -- https://devzone.nordicsemi.com/f/nordic-q-a/5844/why-is-uuid-in-wrong-order
    -- https://devzone.nordicsemi.com/f/nordic-q-a/7306/128-bit-uuid-byte-swap-problem


    If you are on a little-endian system and receive the UUID as per RFC-4122 standard
    (big-endian representation) then yes you'll have to swap bytes on ``time_low``, ``time_mid``,
    and ``time_high_and_version`` in order to use them properly.
    -- https://stackoverflow.com/questions/37711358/how-to-convert-uuid-to-big-endian#comment62896909_37711358


    RFC 4122 may be perfectly clear, but I believe the freedom it permits could cause confusion.
    Here is an example where the default encoding described in RFC 4122 was not adopted:

        Although RFC 4122 recommends network byte order (big-endian) for all fields,
        the PC industry (including the ACPI, UEFI, and Microsoft specifications) has
        consistently used little-endian byte encoding for the first three fields:
        ``time_low``, ``time_mid``, ``time_hi_and_version``. The same encoding,
        also known as wire format, should also be used for the SMBIOS representation
        of the UUID.

    -- http://howtowriteaprogram.blogspot.com/2009/03/uuid-and-byte-order.html


    Background I
    ============
    From Android 4.2, google use Bluedroid stack as its default Bluetooth host stack,
    before android 4.2, its default Bluetooth host stack was Bluez, which is also the
    Linux distribution’s default stack.

    Espressif also incorporated the Bluedroid stack into its ESP-IDF SDK.
    -- https://www.espressif.com/sites/default/files/documentation/esp32_bluetooth_architecture_en.pdf


    Background II
    =============
    Even though Android is the world’s most popular mobile platform, it still
    runs into peculiar problems of non-discoverability of Bluetooth service
    when connected to servers other than the Android server.

    Consider a scenario where a Bluetooth device is configured as a service by
    SDP and Universally Unique Identifier (UUID) is advertised in some order.
    Now, when you start discovering this service on Windows as a discovery
    agent, it will read the configuration in the correct order.

    However, the same action will not happen if the discovery agent is run from
    Android as the discovered UUID gets reversed to what has been advertised.
    For example, Bluetooth service registered on Linux OS with
    UUID as e8e10f95-1a70-4b27-9ccf-02010264e9c8 will be discovered on
    Android OS as c8e96402-0102-cf9c-274b-701a950fe1e8.

    This reversal is done at the hardware level or by the Android Bluetooth SDK—BlueDroid.
    -- https://www.kelltontech.com/kellton-tech-blog/bluetooth-and-its-intricacies-android


    Mitigations
    ===========
    Others also do appropriate workaround by byte-swapping the UUIDs.
    https://gist.github.com/masterjefferson/10922165432ec016a823e46c6eb382e6
    """

    @staticmethod
    def to_bytes_le(uuid):
        """
        Taken from ``uuid2bytes`` from the Pycom user forum. By @jmarcelino.

        https://forum.pycom.io/topic/530/working-with-uuid/2
        """
        uuid = uuid.encode().replace(b'-', b'')
        tmp = ubinascii.unhexlify(uuid)
        return bytes(reversed(tmp))

    @staticmethod
    def from_bytes_le(thing):
        """
        Taken from Pycopy's UUID class. By Paul Sokolovsky.

        https://github.com/pfalcon/pycopy-lib/blob/master/uuid/uuid.py#L5-L20
        """
        if len(thing) != 16:
            raise ValueError('bytes arg must be 16 bytes long')
        hex = ubinascii.hexlify(bytes(reversed(thing))).decode()
        uuid = '-'.join((hex[0:8], hex[8:12], hex[12:16], hex[16:20], hex[20:32]))
        return uuid


def float64(value):
    """

    :param value:

    """
    # https://arduino.stackexchange.com/questions/30502/writing-a-float-to-a-ble-characteristic-what-data-format
    # https://stackoverflow.com/questions/36655172/how-to-read-a-ble-characteristic-float-in-swift
    # Remark: Android-nRF-Connect does not decode float64 values out of the box.
    return pack('<d', value)


def sint16(value):
    """

    :param value:

    """
    # https://github.com/ukBaz/python-bluezero/blob/master/examples/cpu_temperature.py#L28-L29
    return pack('<h', int(value))
    return int(value * 10).to_bytes(2, 'little', True)


def uint16(value):
    """

    :param value:

    """
    return pack('<H', int(value))
    return int(value * 100).to_bytes(2, 'little', True)


def encode_temperature_2a1c(value):
    """Temperature Measurement
    UUID: 2A1C
    Format: Variable length structure
    Value Format: FLOAT (IEEE-11073 32-bit FLOAT)
    Abstract: The Temperature Measurement characteristic is a variable length structure
    containing a Flags field, a Temperature Measurement Value field and, based upon the
    contents of the Flags field, optionally a Time Stamp field and/or a Temperature Type field.
    https://www.bluetooth.com/wp-content/uploads/Sitecore-Media-Library/Gatt/Xml/Characteristics/org.bluetooth.characteristic.temperature_measurement.xml

    IEEE-11073
    ==========
    - https://learn.adafruit.com/introducing-the-adafruit-bluefruit-le-uart-friend/ble-gatt
    - https://devzone.nordicsemi.com/f/nordic-q-a/28727/32-bit-ieee-11073-float-data-type-parser
    - https://stackoverflow.com/questions/11564270/how-to-convert-ieee-11073-16-bit-sfloat-to-simple-float-in-java/14707658#14707658
    - http://www.java2s.com/example/java/language-basics/reads-a-decimal-value-as-a-ieee11073-16bits-float-from-bytebuffer.html
    - https://stackoverflow.com/questions/28899195/converting-two-bytes-to-an-ieee-11073-16-bit-sfloat-in-c-sharp/32950340#32950340

    :param value:

    """
    return bytearray([0b00000000]) + encode_ieee11073(value, precision=5)


def encode_ieee11073(value, precision=2):
    """Binary representation of float value as IEEE-11073:20601 32-bit FLOAT for
    implementing the BLE GATT "Temperature Measurement 2A1C" characteristic.

    -- https://community.hiveeyes.org/t/convenient-ble-gatts-ess-with-micropython/2413/3

    print('Adding Temperature Measurement')
    payload = bytearray([0b00000000]) + float_ieee11073(42.42)
    service.characteristic(uuid=0x2A1C, value=payload)

    :param value:
    :param precision:  (Default value = 2)

    >>> ubinascii.hexlify(encode_ieee11073(42.42))
    b'921000fe'
    """

    # FIXME: Sanity checks dearly required.
    return int(value * (10 ** precision)).to_bytes(3, 'little', True) + pack('<b', -precision)


def decode_ieee11073(payload):
    """

    :param payload:

    >>> decode_ieee11073(encode_ieee11073(167.123456789, precision=5))
    167.1234
    """
    return int.from_bytes(payload[:3], 'little') / 10 ** -unpack('<b', payload[3:])[0]



# -*- coding: utf-8 -*-
# (c) 2019 Richard Pobering <richard@hiveeyes.org>
# (c) 2019 Andreas Motl <andreas@hiveeyes.org>
# License: GNU General Public License, Version 3
from terkin import logging

log = logging.getLogger(__name__)


class UdpServer:
    """ """

    def __init__(self, ip, port):
        self.ip = ip
        self.port = port
        self.callback = None

        self.server_socket = None
        self.is_running = False

    def start(self, callback=None):
        """

        :param callback:  (Default value = None)

        """
        self.callback = callback
        import _thread
        self.is_running = True
        _thread.start_new_thread(self.start_real, ())

    def stop(self):
        self.is_running = False
        log.info("Shutting down UdpServer")
        self.server_socket.close()

    def start_real(self):
        """ """
        import socket

        log.info("Starting UdpServer on {}:{}".format(self.ip, self.port))
        try:
            self.server_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            self.server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.server_socket.bind((self.ip, self.port))

        except Exception as ex:
            log.exc(ex, "Failed starting UdpServer on {}:{}".format(self.ip, self.port))
            return

        try:
            while self.is_running:
                data, addr = self.server_socket.recvfrom(1024)
                self.receive_handler(data, addr)
                self.server_socket.sendto(data, addr)

        except KeyboardInterrupt:
            log.info("Received KeyboardInterrupt within UdpServer")
            self.stop()
            raise

    def receive_handler(self, data, addr):
        """

        :param data: 
        :param addr: 

        """
        log.info('UdpServer received {} from {}'.format(data, addr))
        if callable(self.callback):
            self.callback(data, addr)



from gps import *
import time


class Gpsd():

    def read(self):
        gpsd = gps(mode=WATCH_ENABLE | WATCH_NEWSTYLE)
        counter = 0
        while True:
            report = gpsd.next()
            lat = 0
            lon = 0
            alt = 0
            if report['class'] == 'TPV':
                lat = getattr(report, 'lat', 0.0)
                lon = getattr(report, 'lon', 0.0)
                alt = getattr(report, 'alt', 'nan')
            time.sleep(0.3)
            counter = counter + 1
            if isinstance(lat, float) and isinstance(lon, float) and (isinstance(alt, int) or isinstance(alt, float)):
                data = {'latitude': float(lat), 'longitude': float(lon), 'altitude': float(alt)}
                return data
            if counter > 100:
                break



# -*- coding: utf-8 -*-
# (c) 2019 Andreas Motl <andreas@hiveeyes.org>
# (c) 2019 Richard Pobering <richard@hiveeyes.org>
# License: GNU General Public License, Version 3
import os
import uos
import time
try:
    import os_path
except ImportError:
    import os.path as os_path
from shutil import copyfileobj
from terkin import logging
from terkin.util import file_remove, file_exists, ensure_directory


log = logging.getLogger(__name__)
#log.setLevel(logging.DEBUG)


def backup_file(filename, backup_path, backup_count):
    """Backup file using a rotation mechanism.

    :param filename: param backup_path:
    :param backup_count: return:
    :param backup_path: 

    """

    # Sanity checks.
    if not file_exists(filename):
        log.info('File {} does not exist, skipping backup'.format(filename))
        return

    #log.info('Preparing backup')
    # Running this here will crash the machine.
    #ensure_directory(backup_path)

    filepath = os_path.join(backup_path, os_path.basename(filename))
    outfile = RotatingFile(filepath, backup_count)

    log.info('Creating backup of {}'.format(filename))
    with open(filename, "r") as f:
        outfile.write(f)


class RotatingFile:
    """
    A rotating file handler like RotatingFileHandler.
    
    From logging.handlers.

    """

    def __init__(self, filename, backup_count=0):
        self.filename = filename
        self.backup_count = backup_count

    def write(self, buffer):
        """
        Remove oldest backup, rotate name and write buffer to file.

        :param buffer: buffer to write

        """
        if self.backup_count:

            # Remove the oldest backup file if it is there.
            old_backupfile = "{}.{}".format(self.filename, self.backup_count)
            #log.info('Removing backup file {}'.format(old_backupfile))
            file_remove(old_backupfile)

            # Rename backup files in between oldest and newest one.
            for i in range(self.backup_count - 1, 0, -1):
                if i < self.backup_count:
                    self.rename_file(
                        "{}.{}".format(self.filename, i),
                        "{}.{}".format(self.filename, i + 1))

            # Rename most recent backup file.
            self.rename_file(self.filename, "{}.{}".format(self.filename, 1))

        # Write new most recent backup file.
        log.info('Writing recent backup to {}'.format(self.filename))
        with open(self.filename, "w") as outfile:
            #log.info('copyfileobj: {} => {}'.format(buffer, outfile))
            copyfileobj(buffer, outfile)

        time.sleep_ms(150)
        uos.sync()

    def rename_file(self, oldfile, newfile):
        """
        Rename fila

        :param oldfile: old file name
        :param newfile: new file name

        """
        #log.info('Renaming backup file {} to {}'.format(oldfile, newfile))
        try:
            os.rename(oldfile, newfile)
        except OSError:
            pass
        time.sleep_ms(5)
        uos.sync()
        time.sleep_ms(5)



# -*- coding: utf-8 -*-
# License: GNU General Public License, Version 3
from terkin import logging
from terkin.sensor import SensorManager, AbstractSensor

log = logging.getLogger(__name__)


def includeme(sensor_manager: SensorManager, sensor_info):
    """
    Create MAX17043 sensor object.

    :param sensor_manager:
    :param sensor_info:

    :return: sensor_object
    """
    sensor_object = MAX17043Sensor(settings=sensor_info)
    sensor_bus = sensor_manager.get_bus_by_name(sensor_info.get('bus'))
    sensor_object.acquire_bus(sensor_bus)

    return sensor_object


class MAX17043Sensor(AbstractSensor):
    """
    A generic MAX17043 sensor component.
    
    Used by terkin/datalogger to register and read() from this sensor.
    start() & read() are mandatory.
    """

    def __init__(self, settings=None):

        # from terkin/sensor/core.py class AbstractSensor
        super().__init__(settings=settings)

        # Can be overwritten by ``.set_address()``.
        self.address = self.settings.get('address', 0x36)

    def start(self):
        """Getting the bus"""
        if self.bus is None:
            raise KeyError("Bus missing for MAX17043")

        # Initialize the hardware driver.
        try:
            from max17043 import DFRobot_MAX17043 as MAX17043
            self.driver = MAX17043(i2c=self.bus.adapter)
            return True

        except Exception as ex:
            log.exc(ex, 'MAX17043 hardware driver failed')

    def read(self):
        """ """

        if self.bus is None or self.driver is None:
            return self.SENSOR_NOT_INITIALIZED

        #log.info('Acquire reading from MAX17043')

        data = {}

        voltage = self.driver.readVoltage()
        percentage = self.driver.readPercentage()

        # Prepare readings.
        values = {
            "voltage": voltage / 1000,
            "percentage": percentage,
        }

        # Build telemetry payload.
        fieldnames = values.keys()
        for name in fieldnames:
            fieldname = self.format_fieldname(name, hex(self.address))
            value = values[name]
            data[fieldname] = value

        if not data:
            log.warning("I2C device {} has no value: {}".format(self.address, data))

        log.debug("I2C data:     {}".format(data))

        return data

    


# -*- coding: utf-8 -*-
# (c) 2019 Richard Pobering <richard@hiveeyes.org>
# (c) 2019 Andreas Motl <andreas@hiveeyes.org>
# License: GNU General Public License, Version 3
from machine import Timer
from terkin import logging
from terkin.sensor.touch import TouchPad

log = logging.getLogger(__name__)


class Button:
    """ """
    def __init__(self, name, location, adapter):
        self.name = name
        self.location = location
        self.adapter = adapter

        self.enabled = True

    @property
    def pin(self):
        """ """
        return self.adapter.pin

    def __str__(self):
        return 'Button {} (pin {}) at {}'.format(self.name, self.pin, self.location)


class ButtonManager:
    """ """

    check_interval_ms = 10

    def __init__(self):

        log.info('Initializing button manager')

        # Initialize the buttons.
        self.buttons = []

        # Enable the alarm to check the status.
        self.alarm = Timer.Alarm(self.check, ms=self.check_interval_ms, periodic=True)

    def setup_touchpad(self, pin, name, location):
        """

        :param pin: 
        :param name: 
        :param location: 

        """
        try:
            button = Button(
                name=name,
                location=location,
                adapter=TouchPad(pin, name, sensitivity=1000, duration=500),
            )
            log.info('Setting up %s', button)
            self.buttons.append(button)
        except Exception as ex:
            log.exc(ex, 'Setting up button on pin {} failed'.format(pin))

    def check(self, alarm):
        """

        :param alarm: 

        """
        for button in self.buttons:
            if not button.enabled:
                continue
            try:
                if button.adapter.is_pressed() and not button.adapter.just_pressed():
                    log.info('{} pressed'.format(button))
                    button.adapter.set_press()
            except Exception as ex:
                button.enabled = False
                log.exc(ex, 'Checking {} failed'.format(button))



# -*- coding: utf-8 -*-
# (c) 2020 Jan Hoffmann <jan.hoffmann@bergamsee.de>
# (c) 2020 Andreas Motl <andreas.motl@terkin.org>
# License: GNU General Public License, Version 3
from terkin import logging
from terkin.sensor import SensorManager, AbstractSensor
from terkin.util import get_platform_info

log = logging.getLogger(__name__)
platform_info = get_platform_info()


def includeme(sensor_manager: SensorManager, sensor_info):
    """
    Create EPSolar sensor object.

    :param sensor_manager:
    :param sensor_info:

    :return: sensor_object
    """
    sensor_object = EPSolarSensor(settings=sensor_info)
    return sensor_object


class EPSolarSensor(AbstractSensor):
    """
    About
    =====
    EPSolar ViewStar PWM charge controller sensor component.

    Supported devices
    =================
    - EPSolar ViewStar VS1024N with TTL232 interface and Tracer-MT-5 protocol.

    Resources
    =========
    - http://solpanelerotillbehor.se/Publications/Viewstar%20prod%20info%20140212.pdf
    - https://github.com/xxv/tracer
    - https://github.com/xxv/tracer/network
    - https://github.com/xxv/tracer/blob/master/docs/Protocol-Tracer-MT-5.pdf

    """

    def __init__(self, settings=None):

        super().__init__(settings=settings)

        # Can be overwritten by ``.set_address()``.
        self.device = settings['device']
        self.driver = None

    def start(self):
        """
        Initialize the hardware driver for reading
        the EPSolar ViewStar PWM Charge Controller.
        """
        log.info('Initializing sensor "EPSolar"')

        try:

            # MicroPython
            if platform_info.vendor in [platform_info.MICROPYTHON.Vanilla, platform_info.MICROPYTHON.Pycom]:
                raise NotImplementedError('EPSolar driver not implemented on MicroPython')

            # CPython
            elif platform_info.vendor == platform_info.MICROPYTHON.RaspberryPi:
                from terkin.lib.epsolar import EPSolar
                self.driver = EPSolar(device=self.device)

            else:
                raise NotImplementedError('EPSolar driver not implemented on this platform')

            return True

        except Exception as ex:
            log.exc(ex, 'EPSolar hardware driver failed')

    def read(self):
        """
        Read data from the EPSolar ViewStar PWM Charge Controller.
        """
        log.info('Reading sensor "EPSolar"')

        # MicroPython
        if platform_info.vendor in [platform_info.MICROPYTHON.Vanilla, platform_info.MICROPYTHON.Pycom]:
            raise NotImplementedError('EPSolar driver not implemented on MicroPython')

        # CPython
        elif platform_info.vendor == platform_info.MICROPYTHON.RaspberryPi:
            serial_data, serial_read = self.driver.read_serial()
            serial_data_converted = self.driver.decode_serial(serial_data)
            serial_data_decoded = self.driver.serial_data_prepare(serial_data_converted)
            #serial_data_prepared = self.driver.serial_data_prepare(serial_data)
            data_raw = serial_data_decoded

        else:
            raise NotImplementedError('EPSolar driver not implemented on this platform')

        data = {}
        for key, value in data_raw.items():
            key = 'epsolar.{}'.format(key)
            data[key] = value

        return data



# -*- coding: utf-8 -*-
# (c) 2017-2018 David Gerber <https://github.com/geda>
# HX711 library for the LoPy.
# https://github.com/geda/hx711-lopy
import utime
from terkin import logging
from machine import Pin, enable_irq, disable_irq, idle
from terkin.util import get_platform_info

log = logging.getLogger(__name__)


class HX711:
    """Baseline driver for the HX711 by David Gerber, with modifications.
    
    https://cdn.sparkfun.com/datasheets/Sensors/ForceFlex/hx711_english.pdf


    """

    def __init__(self, dout, pd_sck, gain=128):

        self.platform_info = get_platform_info()

        # Define two pins for clock and data.
        if self.platform_info.vendor == self.platform_info.MICROPYTHON.Vanilla:
            self.pSCK = Pin(int(pd_sck[1:]), mode=Pin.OUT)
            self.pOUT = Pin(int(dout[1:]), mode=Pin.IN, pull=Pin.PULL_UP)
        elif self.platform_info.vendor == self.platform_info.MICROPYTHON.Pycom:
            self.pSCK = Pin(pd_sck, mode=Pin.OUT)
            self.pOUT = Pin(dout, mode=Pin.IN, pull=Pin.PULL_UP)
        else:
            raise NotImplementedError('HX711 is not implemented on this platform')

        self.initialized = False

        self.GAIN = 0
        self.OFFSET = 0
        self.SCALE = 1

        self.time_constant = 0.1
        self.filtered = None

        self.set_gain(gain)

    def set_gain(self, gain):
        """

        :param gain: 

        """
        if gain == 128:
            self.GAIN = 1
        elif gain == 64:
            self.GAIN = 3
        elif gain == 32:
            self.GAIN = 2

    def is_ready(self):
        """ """
        return self.pOUT() == 0

    def initialize(self):
        """Power and initialize the chip.
        Wait for becoming ready.


        """
        if not self.initialized:

            # Wake up the HX711.
            self.power_up()

            # Wait for device to become ready.
            log.info('Initialization started')
            self.wait_ready()
            self.initialized = True

            log.info('Initialization succeeded')
            return True

    def wait_ready(self):
        """ """
        retries = 10000
        #log.info('Waiting for device to become ready')
        while not self.is_ready():
            idle()
            retries -= 1
            if retries == 0:
                raise DeviceNotFound('HX711 not ready')
        #log.info('Device ready')

    def read(self):
        """This chip has a non-standard serial protocol.
        
        Serial Interface
        ----------------
        Pin PD_SCK and DOUT are used for data retrieval, input selection,
        gain selection and power down controls.
        
        When output data is not ready for retrieval, digital output pin DOUT
        is high. Serial clock input PD_SCK should be low. When DOUT goes to
        low, it indicates data is ready for retrieval.
        
        By applying 25~27 positive clock pulses at the PD_SCK pin, data is
        shifted out from the DOUT output pin. Each PD_SCK pulse shifts out
        one bit, starting with the MSB bit first, until all 24 bits are
        shifted out. The 25th pulse at PD_SCK input will pull DOUT pin back
        to high.


        """

        # Initialize the hardware once.
        # Otherwise, croak with ``DeviceNotFound('HX711 not available')``.
        if not self.initialize():

            # Wait for the device becoming ready.
            self.wait_ready()

        # Shift in data, gain & channel info.
        result = 0
        for j in range(24 + self.GAIN):
            state = disable_irq()
            self.pSCK(True)
            self.pSCK(False)
            enable_irq(state)
            result = (result << 1) | self.pOUT()

        # Shift back the extra bits.
        result >>= self.GAIN

        # Check sign.
        if result > 0x7fffff:
            result -= 0x1000000

        return result

    def read_average(self, times=3):
        """

        :param times:  (Default value = 3)

        """
        sum = 0
        for i in range(times):
            sum += self.read()
        return sum / times

    def read_lowpass(self):
        """ """
        if self.filtered is None:
            self.filtered = self.read()
        self.filtered += self.time_constant * (self.read() - self.filtered)
        return self.filtered

    def get_value(self, times=3):
        """

        :param times:  (Default value = 3)

        """
        return self.read_average(times) - self.OFFSET

    def get_units(self, times=3):
        """

        :param times:  (Default value = 3)

        """
        return self.get_value(times) / self.SCALE

    def tare(self, times=15):
        """

        :param times:  (Default value = 15)

        """
        sum = self.read_average(times)
        self.set_offset(sum)

    def set_scale(self, scale):
        """

        :param scale: 

        """
        self.SCALE = scale

    def set_offset(self, offset):
        """

        :param offset: 

        """
        self.OFFSET = offset

    def set_time_constant(self, time_constant = None):
        """

        :param time_constant:  (Default value = None)

        """
        if time_constant is None:
            return self.time_constant
        elif 0 < time_constant < 1.0:
            self.time_constant = time_constant

    def power_up(self):
        """When PD_SCK Input is low, chip is in normal working mode."""

        # Unfreeze pin hold when coming from deep sleep.
        # https://community.hiveeyes.org/t/strom-sparen-beim-einsatz-der-micropython-firmware-im-batteriebetrieb/2055/72
        if self.platform_info.vendor == self.platform_info.MICROPYTHON.Pycom:
            self.pSCK.hold(False)

        log.info('HX711 power up')
        self.pSCK.value(False)
        #utime.sleep_us(80)

        #self.initialize()

    def power_down(self):
        """When PD_SCK pin changes from low to high and stays at
        high for longer than 60µs, HX711 enters power down mode.


        """
        log.info('HX711 power down')
        state = disable_irq()
        self.pSCK.value(False)
        self.pSCK.value(True)
        utime.sleep_us(80)
        enable_irq(state)

        # Hold level to HIGH, even during deep sleep.
        # https://community.hiveeyes.org/t/strom-sparen-beim-einsatz-der-micropython-firmware-im-batteriebetrieb/2055/72
        if self.platform_info.vendor == self.platform_info.MICROPYTHON.Pycom:
            self.pSCK.hold(True)


class DeviceNotFound(Exception):
    """ """
    pass



# -*- coding: utf-8 -*-
# (c) 2019-2020 Andreas Motl <andreas@hiveeyes.org>
# (c) 2019-2020 Richard Pobering <richard@hiveeyes.org>
# (c) 2019-2020 Jan Hoffmann <jan.hoffmann@bergamsee.de>
# License: GNU General Public License, Version 3
import time
import machine

from __main__ import bootloader
from umal import ApplicationInfo, PlatformInfo

from terkin import __version__
from terkin import logging
from terkin.exception import SensorUnknownError
from terkin.configuration import TerkinConfiguration
from terkin.device import TerkinDevice
from terkin.network import SystemWiFiMetrics
from terkin.sensor import SensorManager, AbstractSensor
from terkin.model import SensorReading, DataFrame
from terkin.sensor.system import SystemMemoryFree, SystemTemperature, SystemVoltage, SystemUptime
from terkin.util import gc_disabled

log = logging.getLogger(__name__)


class TransientStorage:
    """
    Store last sensor reading for serving through HTTP API.
    """

    def __init__(self):
        self.last_reading = {}


# Maybe refactor to TerkinCore.
class TerkinDatalogger:
    """ 
    Main class of project.
    Handles loop & sleep, registers sensors, reads their data and stores them.
    Shows up as 'datalogger' in the rest of the program.
    """

    # Application metadata.
    name = 'Terkin Datalogger'
    version = __version__

    # For the singleton factory.
    __instance__ = None

    def __init__(self, settings, platform_info: PlatformInfo = None):

        # Reference to the chronometer used for general timekeeping.
        self.duty_chrono = bootloader.duty_chrono

        # Signal startup with first available timestamp.
        log.info('Starting Terkin datalogger')

        # Obtain configuration settings.
        self.settings = TerkinConfiguration()
        self.settings.add(settings)
        self.settings.add_user_file()

        # Configure logging.
        logging_enabled = self.settings.get('main.logging.enabled', False)
        if not logging_enabled:
            log.info('Disabling logging to save bytes')
            logging.disable_logging()

        # Initialize ApplicationInfo object.
        self.application_info = ApplicationInfo(
            name=self.name, version=self.version, settings=self.settings,
            application=self, platform_info=platform_info)

        # Initialize transient storage.
        self.storage = TransientStorage()

        # Initialize device.
        self.device = TerkinDevice(self.application_info)

        # Button manager instance (optional).
        self.button_manager = None

        # Initialize sensor domain.
        self.sensor_manager = SensorManager(self.settings)

    def setup(self):

        # Report about wakeup reason and run wakeup tasks.
        self.device.resume()

        # Start the watchdog for sanity.
        self.device.watchdog.start()

        # Configure RGB-LED according to settings.
        self.device.configure_rgb_led()

        # Alternative startup signalling: 2 x green.
        self.device.blink_led(0x000b00, count=2)

        # Free up some memory.
        self.device.run_gc()

        # Turn off LTE modem and Bluetooth as we don't use them yet.
        # TODO: Make this configurable.
        has_lte_modem = self.application_info.platform_info.device_name in ['GPy', 'FiPy']
        lte_enabled = self.settings.get('networking.lte.enabled')
        if has_lte_modem and not lte_enabled:
            self.device.power_off_lte_modem()

        self.device.power_off_bluetooth()

        log.info('Starting %s', self.application_info.fullname)

        # Dump configuration settings.
        log_configuration = self.settings.get('main.logging.configuration', False)
        if log_configuration:
            self.settings.dump()

        # Disable this if you don't want serial access.
        #self.device.enable_serial()

        # Hello world.
        self.device.print_bootscreen()

        # Start networking and telemetry subsystems.

        # Conditionally start network services and telemetry if networking is available.
        try:
            self.device.start_networking()
        except Exception as ex:
            log.exc(ex, 'Networking subsystem failed')
            self.device.status.networking = False

        self.device.start_telemetry()

        # Todo: Signal readyness by publishing information about the device (Microhomie).
        # e.g. ``self.device.publish_properties()``

        self.setup_sensors()

    def setup_sensors(self):

        # Setup sensors.
        log.info('Setting up sensors')
        self.device.watchdog.feed()
        bus_settings = self.settings.get('sensors.buses', self.settings.get('sensors.busses', []))
        self.sensor_manager.setup_buses(bus_settings)
        self.register_sensors()
        self.sensor_manager.start_sensors()

        log.info('Setup finished')

    def start(self):
        self.start_mainloop()

    def start_mainloop(self):
        """ """

        # Todo: Refactor by using timers.

        # Enter the main loop.
        while True:
            self.duty_task()

    def duty_task(self):
        """Main duty cycle task."""

        # Feed the watchdog timer to keep the system alive.
        self.device.watchdog.feed()

        # Indicate activity.
        # Todo: Optionally disable this output.
        log.info('--- cycle ---')

        # Run downstream mainloop handlers.
        self.duty_cycle()

        # Sleep how configured
        self.sleep()

    def duty_cycle(self):
        """Main duty cycle"""

        if not self.settings.get('main.deepsleep', False):
            self.duty_chrono.reset()

        #log.info('Terkin loop')

        # Alternative loop signalling: 1 x blue.
        # https://forum.pycom.io/topic/2067/brightness-of-on-board-led/7
        self.device.blink_led(0x00000b, count=2)

        # Read sensors.
        readings = self.read_sensors()

        # Remember current reading
        self.storage.last_reading = readings.data_in

        # Run the garbage collector.
        self.device.run_gc()

        # Transmit data.
        transmission_success = self.transmit_readings(readings)

        # Signal transmission outcome.
        if transmission_success:
            self.device.blink_led(0x00000b)
        else:
            self.device.blink_led(0x0b0000)

        # Run the garbage collector.
        self.device.run_gc()

        # Give the system some breath.
        machine.idle()

    def sleep(self):
        """Sleep or shutoff until the next measurement cycle."""

        lightsleep = self.settings.get('main.lightsleep', False)
        deepsleep = self.settings.get('main.deepsleep', False)
        shutoff = self.settings.get('main.shutoff', False)
        interval = self.get_sleep_time()

        # Amend deep sleep intent when masked through maintenance mode.
        if self.device.status.maintenance is True:
            lightsleep = False
            deepsleep = False
            shutoff = False
            log.info('Device is in maintenance mode. Skipping deep sleep and '
                     'adjusting sleep time to {} seconds.'.format(interval))

        # Prepare device shutdown.
        try:
            # Shut down sensor peripherals.
            self.sensor_manager.power_off()

            # Shut down networking.
            if deepsleep or shutoff:
                self.device.networking.stop()

        except Exception as ex:
            log.exc(ex, 'Power off failed')

        if shutoff:
            # shut off the MCU via DS3231
            self.shutoff()
        else:
            # Activate device sleep mode.
            try:
                self.device.hibernate(interval, lightsleep=lightsleep, deepsleep=deepsleep)

            # When hibernation fails, fall back to regular "time.sleep".
            except Exception as ex:
                log.exc(ex, 'Failed to hibernate, falling back to regular sleep')
                # Todo: Emit error message here.
                log.info('Sleeping for {} seconds'.format(interval))
                time.sleep(interval)

    def get_sleep_time(self):
        """ 
        Calculate the next sleep intervall.
        """

        interval = self.settings.get('main.interval', 60.0)

        # Configuration switchover backward compatibility / defaults.
        if isinstance(interval, (float, int)):
            self.settings.set('main.interval', {})
            self.settings.setdefault('main.interval.field', interval)
        self.settings.setdefault('main.interval.maintenance', 5.0)

        interval = self.settings.get('main.interval.field')

        # First, try to acquire deep sleep interval from NVRAM.
        # This gets used when set from a LoRaWAN downlink message.
        # pycom.nvs_get should return "None" in case of unset key. Instead it throws an error
        try:
            import pycom
            interval_minutes = pycom.nvs_get('deepsleep')
            if isinstance(interval_minutes, int):
                log.info('Deep sleep interval set to %s minute(s) by LoRaWAN downlink message', interval_minutes)
                interval = interval_minutes * 60

        # Otherwise, use original configuration setting.
        except Exception as ex:
            pass

        # Amend deep sleep intent when masked through maintenance mode.
        if self.device.status.maintenance is True:
            interval = self.settings.get('main.interval.maintenance')

        # Compute sleeping duration from measurement interval and elapsed time.
        elapsed = self.duty_chrono.read()
        sleep_time = interval - elapsed

        if sleep_time <= 0:
            sleep_time = interval

        return sleep_time

    def register_sensors(self):
        """
        Configure and register sensor objects.
        There are three types of sensors: system, environment & buses.

        The sensors are registered by calling their respective classes
        from terkin/driver.

        Definitions are in 'settings.py'.
        """

        # Add sensors.
        log.info('Registering sensors')
        sensor_infos = []

        # Get list of system sensors from configuration settings.
        sensor_infos += self.settings.get('sensors.system', [])

        # Get list of environmental sensors from configuration settings.
        sensor_infos += self.settings.get('sensors.environment', [])

        # Backward compatibility for environmental sensors.
        if sensor_infos is None:
            sensor_infos += self.settings.get('sensors.registry', {}).values() or []

        # Scan sensor definitions, create and register sensor objects.
        for sensor_info in sensor_infos:

            sensor_type = sensor_info.get('type', 'unknown').lower()
            sensor_id = sensor_info.get('id', sensor_info.get('key', sensor_type))

            # Skip sensor if disabled in configuration.
            if not sensor_info.get('enabled', False):
                log.debug('Sensor with id={} and type={} is disabled, skipping registration'.format(sensor_id, sensor_type))
                continue

            # Skip WiFi sensor registration when WiFi is disabled.
            if sensor_type == 'system.wifi':
                if not self.settings.get('networking.wifi.enabled'):
                    log.info('WiFi is disabled, skipping sensor registration')
                    continue

            self.register_sensor(sensor_info)

            # Clean up memory after creating each sensor object.
            #self.device.run_gc()

    def register_sensor(self, sensor_info):
        """
        Register one sensor.
        """
        sensor_type = sensor_info.get('type', 'unknown').lower()
        sensor_id = sensor_info.get('id', sensor_info.get('key', sensor_type))
        description = sensor_info.get('description')

        # Resolve associated bus object.
        sensor_bus = None
        sensor_bus_name = None
        if 'bus' in sensor_info:
            sensor_info_bus = sensor_info['bus']
            sensor_bus = self.sensor_manager.get_bus_by_name(sensor_info_bus)

            # Skip sensor if associated bus is disabled in configuration.
            if sensor_bus is None:
                log.info('Bus {} for sensor with id={} and type={} is disabled, '
                         'skipping registration'.format(sensor_info_bus, sensor_id, sensor_type))
                return
            sensor_bus_name = sensor_bus.name

        # Human readable sensor address.
        if 'address' in sensor_info:
            sensor_address = hex(sensor_info.get('address'))
        else:
            sensor_address = None

        # Report sensor registration to user.
        message = 'Setting up sensor with id={} and type={} on bus={} with address={}'.format(
            sensor_id, sensor_type, sensor_bus_name, sensor_address)
        if description:
            message += ' described as "{}"'.format(description)
        log.info(message)

        # Backward compat.
        if sensor_type == 'ds18b20':
            sensor_type = 'ds18x20'

        # Registration NG
        # Run self-registration procedure by invoking
        # the "includeme()" function on each sensor module.
        try:

            # Load sensor module.
            import terkin.driver
            modulename = '{}_sensor'.format(sensor_type)
            fullname = 'terkin.driver.{}'.format(modulename)
            log.info('Importing module "{}"'.format(fullname))
            __import__(fullname)
            module = getattr(terkin.driver, modulename)

            # Acquire sensor object.
            includeme = getattr(module, 'includeme')
            sensor_object = includeme(self.sensor_manager, sensor_info)

            # Register sensor with sensor manager.
            self.sensor_manager.register_sensor(sensor_object)

            return

        except ImportError as ex:
            if not fullname.startswith('terkin.driver.system'):
                log.error('Driver module "{}" not found'.format(fullname))

        except AttributeError as ex:
            if "has no attribute 'includeme'" in str(ex):
                log.warning('Driver module "{}" is deprecated, "includeme" is missing'.format(fullname))
            else:
                log.exc(ex, 'Driver module "{}" failed'.format(fullname))

        except Exception as ex:
            log.exc(ex, 'Registering driver module "{}" failed'.format(fullname))

        # Legacy registration
        try:
            self.register_sensor_legacy(sensor_info, sensor_bus)
        except Exception as ex:
            log.exc(ex, 'Setting up sensor with id={} and type={} failed'.format(sensor_id, sensor_type))

    def register_sensor_legacy(self, sensor_info, sensor_bus):

        sensor_type = sensor_info.get('type', 'unknown').lower()

        # Sensor reporting about free system memory.
        if sensor_type == 'system.memfree':
            sensor_object = SystemMemoryFree(sensor_info)

        # Sensor which reports system temperature.
        elif sensor_type == 'system.temperature':
            sensor_object = SystemTemperature(sensor_info)

        # Sensor which reports battery voltage.
        elif sensor_type in ['system.voltage.battery', 'system.battery-voltage']:
            sensor_object = SystemVoltage(sensor_info)

        # Sensor which reports solar panel voltage.
        elif sensor_type == 'system.voltage.solar':
            sensor_object = SystemVoltage(sensor_info)

        # Sensor which reports system uptime metrics.
        elif sensor_type == 'system.uptime':
            sensor_object = SystemUptime(sensor_info)

        # Sensor which reports WiFi metrics.
        elif sensor_type == 'system.wifi':
            try:
                sensor_object = SystemWiFiMetrics(sensor_info, self.device.networking.wifi_manager.station)
            except Exception as ex:
                log.exc(ex, 'Enabling SystemWiFiMetrics sensor failed')
                return

        # Initialize buttons / touch pads.
        elif sensor_type == 'system.touch-buttons':
            from terkin.sensor.button import ButtonManager
            self.button_manager = ButtonManager()
            self.start_buttons()
            return

        else:
            raise SensorUnknownError('Unknown sensor type "{}"'.format(sensor_type))

        # Register sensor object with sensor manager.
        self.sensor_manager.register_sensor(sensor_object)

    def read_sensors(self) -> DataFrame:
        """
        Read measurements from all sensor objects that have been registered in the sensor_manager.
        Reading is done with the read() function of each respective sensor object.
        """

        # Power up sensor peripherals.
        self.sensor_manager.power_on()

        # Collect observations.
        data = {}
        richdata = {}
        readings = []

        # Iterate all registered sensors.
        sensors = self.sensor_manager.sensors
        log.info('Reading %s sensor ports', len(sensors))
        for sensor in sensors:

            # Signal sensor reading to user.
            sensorname = sensor.__class__.__name__
            log.info('Reading sensor port "%s"', sensorname)

            # Read sensor port.
            try:

                # Disable garbage collector to guarantee reasonable
                # realtime behavior before invoking sensor reading.
                with gc_disabled():
                    sensor_outcome = sensor.read()

                # Power off HX711 after reading
                if "HX711Sensor" in sensorname:
                   sensor.power_off()

                # Backward compat.
                if isinstance(sensor_outcome, SensorReading):
                    sensor_reading = sensor_outcome
                else:
                    sensor_reading = SensorReading()
                    sensor_reading.sensor = sensor
                    sensor_reading.data = sensor_outcome

                sensor_data = sensor_reading.data

                # Evaluate sensor outcome.
                if sensor_data is None or sensor_data is AbstractSensor.SENSOR_NOT_INITIALIZED:
                    continue

                # Round values according to sensor settings.
                if sensor.settings.get('decimals') is not None:
                    for key, value in sensor_data.items():
                        sensor_data[key] = round(sensor_data[key], sensor.settings.get('decimals'))

                # Add sensor reading to observations.
                data.update(sensor_data)

                # Record reading for prettified output.
                self.record_reading(sensor, sensor_data, richdata)

                readings.append(sensor_reading)

            except Exception as ex:
                # Because of the ``gc_disabled`` context manager used above,
                # the propagation of exceptions has to be tweaked like that.
                log.exc(ex, 'Reading sensor "%s" failed', sensorname)

            # Feed the watchdog.
            self.device.watchdog.feed()

            # Clean up memory after reading each sensor object.
            #self.device.run_gc()

        # Debugging: Print sensor data before running telemetry.
        prettify_log = self.settings.get('sensors.prettify_log', False)
        if prettify_log:
            from terkin.util import ddformat
            log.info('Sensor data:\n\n%s', ddformat(richdata, indent=11))
        else:
            log.info('Sensor data:  %s', data)

        # Capture all sensor readings.
        result = DataFrame()
        result.readings = readings
        result.data_in = data

        return result

    def record_reading(self, sensor, reading, richdata):
        """

        :param sensor: 
        :param reading: 
        :param richdata: 

        """
        for key, value in reading.items():
            richdata[key] = {'value': value}
            if hasattr(sensor, 'settings') and 'description' in sensor.settings:
                richdata[key]['description'] = sensor.settings.get('description')
                # Hack to propagate the correct detail-description to prettified output.
                # TODO: Attach settings directly to its reading, while actually reading it.
                if 'devices' in sensor.settings:
                    for device_settings in sensor.settings['devices']:
                        device_address = device_settings['address'].lower()
                        if device_address in key:
                            if hasattr(sensor, 'get_device_description'):
                                device_description = sensor.get_device_description(device_address)
                                if device_description:
                                    richdata[key]['description'] = device_description

    def transmit_readings(self, dataframe: DataFrame):
        """
        Transmit data

        :param dataframe:

        """

        # TODO: Optionally disable telemetry.
        if self.device.telemetry is None:
            log.warning('Telemetry disabled')
            return False

        telemetry_status = self.device.telemetry.transmit(dataframe)
        count_total = len(telemetry_status)
        success = all(telemetry_status.values())

        # Evaluate telemetry status outcome.
        if success:
            log.info('Telemetry status: SUCCESS ({}/{})'.format(count_total, count_total))
        else:
            count_failed = len([item for item in telemetry_status.values() if item is not True])
            log.warning('Telemetry status: FAILURE. {} out of {} targets failed. '
                        'Status: {}'.format(count_failed, count_total, telemetry_status))

        return success

    def start_buttons(self):

        """
        Configure ESP32 touchpads.
        """

        # RGB-LED: 2
        # POWER-ENABLE: 3
        # SD-Card: 4, 8
        # LTE 19, 20
        # Misc: 13, 14, 9, 23

        # Physical location when looking at the board with the RGB-LED oriented to the top.

        # Location: Left side, 6th pin from top.
        self.button_manager.setup_touchpad('P4', name='Touch3', location='Module-Left-Top-6th')

        # Location: Left side, 5th pin from bottom.
        self.button_manager.setup_touchpad('P8', name='Touch2', location='Module-Left-Bottom-5th')

        # Location: Right side.
        self.button_manager.setup_touchpad('P23', name='Touch6', location='Module-Right-Top-4th')

        # Location: Right side.
        # ValueError: invalid pin for touchpad
        """
        P18 and P17 are able to wake up on rising and falling edge. These two pins have internal
        pull-ups configurable by software (Pull-downs if needed must be added externally)

        -- https://docs.pycom.io/gitbook/assets/deepsleep-pinout.pdf


        ext0 External Wake-up Source
        RTC controller contains logic to trigger wake-up when one particular pin is set to
        a predefined logic level. That pin can be one of RTC GPIOs 0,2,4,12-15,25-27,32-39.

        -- https://lastminuteengineers.com/esp32-deep-sleep-wakeup-sources/#ext0-external-wakeup-source

        """
        #self.button_manager.setup_touchpad('P17', name='TouchX', location='Module-Right-Bottom-5th')
        #self.button_manager.setup_touchpad('P18', name='TouchY', location='Module-Right-Bottom-6th')

        # Will yield ``ValueError: Touch pad error``.
        #self.button_manager.setup_touchpad('P20', name='Touch8', location='Module-Right-Top-7th')
        #self.button_manager.setup_touchpad('P19', name='Touch9', location='Module-Right-Top-8th')

    def shutoff(self):
        """ shut off the MCU """

        import DS3231tokei
        import utime
        from machine import Pin, RTC

        bus = self.sensor_manager.get_bus_by_sensortype('DS3231')
        ds = DS3231tokei.DS3231(bus.adapter)
        interval = self.settings.get('main.interval.shutoff', 10) * 60  # convert from minutes to seconds
        (year,month,day,dotw,hour,minute,second) = ds.getDateTime() # get the current time

        rtc = RTC() # create RTC
        if year < 2001:
            year = 2001 # sanity check, as of mpy 1.12 year must be >= 2001
        rtc.init((year,month,day,dotw,hour,minute,second,0)) # set time

        # check if its night or winter and adjust interval
        night_start = self.settings.get('main.interval.night_start', 0)
        night_end =  self.settings.get('main.interval.night_end', 0)
        winter_start = self.settings.get('main.interval.winter_start', 0)
        winter_end =  self.settings.get('main.interval.winter_end', 0)
        if night_start > 0 and (hour >= night_start or hour <= night_end):  # double interval for the night
            interval *= 2
        if winter_start > 0 and (month >= winter_start or month <= winter_end):  # double interval for winter
            interval *= 2

        # Compute sleeping duration from measurement interval and elapsed time.
        elapsed = int(self.duty_chrono.read())
        now_secs = utime.mktime(utime.localtime())
        wake_at = now_secs - elapsed + interval
        if (wake_at - now_secs) < 180:  # don't shutoff for less than 3 minutes
            wake_at += interval

        (year,month,day,hour,minute,second, dotw, doty) = utime.localtime(wake_at) # convert the wake up time

        # set alarm
        ds.setAlarm2(day,hour,minute, DS3231tokei.A2_ON_HOUR_MINUTE)

        # turn off MCU via MOSFET
        utime.sleep(1)
        ds.enableAlarm2()
        ds.resetAlarm2()        

        # The End

    def scale_wizard(self):
        """
        Invoke scale adjustment wizard.

        Synopsis:

        - Invoke Terkin datalogger.
        - Interrupt by pressing CTRL+C.
        - Type ``datalogger.scale_wizard()``.
        """

        # Setup sensors.
        self.setup_sensors()

        # Invoke scale adjustment routine.
        from terkin.sensor.scale import ScaleAdjustment
        adj = ScaleAdjustment(sensor_manager=self.sensor_manager)
        adj.start_wizard()

